import asyncio
import re
from decimal import Decimal
from typing import Any, Dict, List, Optional, Set, Union

from async_timeout import timeout

from hummingbot.connector.gateway.amm.gateway_amm_base import GatewayAMMBase
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder
from hummingbot.core.data_type.cancellation_result import CancellationResult
from hummingbot.core.data_type.in_flight_order import OrderState, OrderUpdate
from hummingbot.core.event.events import (
    TokenApprovalCancelledEvent,
    TokenApprovalEvent,
    TokenApprovalFailureEvent,
    TokenApprovalSuccessEvent,
    TradeType,
)
from hummingbot.core.gateway import check_transaction_exceptions
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather

s_logger = None
s_decimal_0 = Decimal("0")
s_decimal_NaN = Decimal("nan")


class GatewayEthereumAMM(GatewayAMMBase):
    """
    Defines Ethereum-specific functions for interacting with AMM protocols via Gateway.
    """

    APPROVAL_ORDER_ID_PATTERN = re.compile(r"approve-(\w+)-(\w+)")

    _ev_loop: asyncio.AbstractEventLoop
    _allowances: Dict[str, Decimal]
    _update_allowances: Optional[asyncio.Task]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._ev_loop = asyncio.get_event_loop()
        self._allowances = {}
        self._update_allowances = None

    async def start_network(self):
        if self._trading_required:
            self._status_polling_task = safe_ensure_future(self._status_polling_loop())
            self._update_allowances = safe_ensure_future(self.update_allowances())
            self._get_gas_estimate_task = safe_ensure_future(self.get_gas_estimate())
        self._get_chain_info_task = safe_ensure_future(self.get_chain_info())

    async def stop_network(self):
        if self._status_polling_task is not None:
            self._status_polling_task.cancel()
            self._status_polling_task = None
        if self._update_allowances is not None:
            self._update_allowances.cancel()
            self._update_allowances = None
        if self._get_chain_info_task is not None:
            self._get_chain_info_task.cancel()
            self._get_chain_info_task = None
        if self._get_gas_estimate_task is not None:
            self._get_gas_estimate_task.cancel()
            self._get_chain_info_task = None

    async def _status_polling_loop(self):
        await self.update_balances(on_interval=False)
        while True:
            try:
                self._poll_notifier = asyncio.Event()
                await self._poll_notifier.wait()
                await safe_gather(
                    self.update_balances(on_interval=True),
                    self.update_canceling_transactions(self.canceling_orders),
                    self.update_token_approval_status(self.approval_orders),
                    self.update_order_status(self.amm_orders)
                )
                self._last_poll_timestamp = self.current_timestamp
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.logger().error(str(e), exc_info=True)

    @property
    def approval_orders(self) -> List[GatewayInFlightOrder]:
        return [
            approval_order
            for approval_order in self._order_tracker.active_orders.values()
            if approval_order.is_approval_request
        ]

    @property
    def canceling_orders(self) -> List[GatewayInFlightOrder]:
        return [
            cancel_order
            for cancel_order in self.amm_orders
            if cancel_order.is_pending_cancel_confirmation
        ]

    def create_approval_order_id(self, token_symbol: str) -> str:
        return f"approve-{self.connector_name}-{token_symbol}"

    def get_token_symbol_from_approval_order_id(self, approval_order_id: str) -> Optional[str]:
        match = self.APPROVAL_ORDER_ID_PATTERN.search(approval_order_id)
        if match:
            return match.group(2)
        return None

    def is_pending_approval(self, token: str) -> bool:
        for order in self.approval_orders:
            if token in order.client_order_id:
                return order.is_pending_approval
        return False

    async def approve_token(self, token_symbol: str, **request_args) -> Optional[GatewayInFlightOrder]:
        """
        Approves contract as a spender for a token.
        :param token_symbol: token to approve.
        """
        approval_id: str = self.create_approval_order_id(token_symbol)

        self.logger().info(f"Initiating approval for {token_symbol}.")

        self.start_tracking_order(order_id=approval_id,
                                  trading_pair=token_symbol,
                                  is_approval=True)
        try:
            resp: Dict[str, Any] = await self._get_gateway_instance().approve_token(
                self.chain,
                self.network,
                self.address,
                token_symbol,
                self.connector_name,
                **request_args
            )

            transaction_hash: Optional[str] = resp.get("approval", {}).get("hash")
            nonce: Optional[int] = resp.get("nonce")
            if transaction_hash is not None and nonce is not None:
                tracked_order = self._order_tracker.fetch_order(client_order_id=approval_id)
                tracked_order.update_exchange_order_id(transaction_hash)
                tracked_order.nonce = nonce
                self.logger().info(
                    f"Maximum {token_symbol} approval for {self.connector_name} contract sent, hash: {transaction_hash}."
                )
                return tracked_order
            else:
                self.stop_tracking_order(approval_id)
                self.logger().info(f"Approval for {token_symbol} on {self.connector_name} failed.")
                return None
        except Exception:
            self.stop_tracking_order(approval_id)
            self.logger().error(
                f"Error submitting approval order for {token_symbol} on {self.connector_name}-{self.network}.",
                exc_info=True
            )
            return None

    async def update_allowances(self):
        """
        Allowances updated continously.
        """
        while True:
            self._allowances = await self.get_allowances()
            await asyncio.sleep(120)  # sleep for 2 mins

    async def get_allowances(self) -> Dict[str, Decimal]:
        """
        Retrieves allowances for token in trading_pairs
        :return: A dictionary of token and its allowance.
        """
        ret_val = {}
        resp: Dict[str, Any] = await self._get_gateway_instance().get_allowances(
            self.chain, self.network, self.address, list(self._tokens), self.connector_name
        )
        for token, amount in resp["approvals"].items():
            ret_val[token] = Decimal(str(amount))
        return ret_val

    def parse_price_response(
        self,
        base: str,
        quote: str,
        amount: Decimal,
        side: TradeType,
        price_response: Dict[str, Any],
        process_exception: bool = False
    ) -> Optional[Decimal]:
        """
        Parses price response
        :param base: The base asset
        :param quote: The quote asset
        :param amount: amount
        :param side: trade side
        :param price_response: Price response from Gateway.
        :param process_exception: Flag to trigger error on exception
        """
        required_items = ["price", "gasLimit", "gasPrice", "gasCost", "gasPriceToken"]
        if any(item not in price_response.keys() for item in required_items):
            if "info" in price_response.keys():
                self.logger().info(f"Unable to get price. {price_response['info']}")
            else:
                self.logger().info(f"Missing data from price result. Incomplete return result for ({price_response.keys()})")
        else:
            gas_price_token: str = price_response["gasPriceToken"]
            gas_cost: Decimal = Decimal(price_response["gasCost"])
            price: Decimal = Decimal(price_response["price"])
            # self.network_transaction_fee = TokenAmount(gas_price_token, gas_cost)
            if process_exception is True:
                gas_limit: int = int(price_response["gasLimit"])
                exceptions: List[str] = check_transaction_exceptions(
                    balances=self._account_balances,
                    base_asset=base,
                    quote_asset=quote,
                    amount=amount,
                    side=side,
                    gas_limit=gas_limit,
                    gas_cost=gas_cost,
                    gas_asset=gas_price_token,
                    swaps_count=len(price_response.get("swaps", [])),
                    allowances=self._allowances
                )
                for index in range(len(exceptions)):
                    self.logger().warning(
                        f"Warning! [{index + 1}/{len(exceptions)}] {side} order - {exceptions[index]}"
                    )
                if len(exceptions) > 0:
                    return None
            return Decimal(str(price))
        return None

    async def update_token_approval_status(self, tracked_approvals: List[GatewayInFlightOrder]):
        """
        Calls REST API to get status update for each in-flight token approval transaction.
        """
        if len(tracked_approvals) < 1:
            return
        tx_hash_list: List[str] = await safe_gather(*[
            tracked_approval.get_exchange_order_id() for tracked_approval in tracked_approvals
        ])
        transaction_states: List[Union[Dict[str, Any], Exception]] = await safe_gather(*[
            self._get_gateway_instance().get_transaction_status(
                self.chain,
                self.network,
                tx_hash
            )
            for tx_hash in tx_hash_list
        ], return_exceptions=True)
        for tracked_approval, transaction_status in zip(tracked_approvals, transaction_states):
            token_symbol: str = self.get_token_symbol_from_approval_order_id(tracked_approval.client_order_id)
            if isinstance(transaction_status, Exception):
                self.logger().error(f"Error while trying to approve token {token_symbol} for {self.connector_name}: "
                                    f"{transaction_status}")
                continue
            if "txHash" not in transaction_status:
                self.logger().error(f"Error while trying to approve token {token_symbol} for {self.connector_name}: "
                                    "txHash key not found in transaction status.")
                continue
            if transaction_status["txStatus"] == 1:
                if transaction_status["txReceipt"]["status"] == 1:
                    self.logger().info(f"Token approval for {tracked_approval.client_order_id} on {self.connector_name} "
                                       f"successful.")
                    tracked_approval.current_state = OrderState.APPROVED
                    self.trigger_event(
                        TokenApprovalEvent.ApprovalSuccessful,
                        TokenApprovalSuccessEvent(
                            self.current_timestamp,
                            self.connector_name,
                            token_symbol
                        )
                    )
                else:
                    self.logger().warning(
                        f"Token approval for {tracked_approval.client_order_id} on {self.connector_name} failed."
                    )
                    tracked_approval.current_state = OrderState.FAILED
                    self.trigger_event(
                        TokenApprovalEvent.ApprovalFailed,
                        TokenApprovalFailureEvent(
                            self.current_timestamp,
                            self.connector_name,
                            token_symbol
                        )
                    )
                self.stop_tracking_order(tracked_approval.client_order_id)

    async def update_canceling_transactions(self, canceled_tracked_orders: List[GatewayInFlightOrder]):
        """
        Update tracked orders that have a cancel_tx_hash.
        :param canceled_tracked_orders: Canceled tracked_orders (cancel_tx_has is not None).
        """
        if len(canceled_tracked_orders) < 1:
            return

        self.logger().debug(
            "Polling for order status updates of %d canceled orders.",
            len(canceled_tracked_orders)
        )
        update_results: List[Union[Dict[str, Any], Exception]] = await safe_gather(*[
            self._get_gateway_instance().get_transaction_status(
                self.chain,
                self.network,
                tx_hash
            )
            for tx_hash in [t.cancel_tx_hash for t in canceled_tracked_orders]
        ], return_exceptions=True)
        for tracked_order, update_result in zip(canceled_tracked_orders, update_results):
            if isinstance(update_result, Exception):
                raise update_result
            if "txHash" not in update_result:
                self.logger().error(f"No txHash field for transaction status of {tracked_order.client_order_id}: "
                                    f"{update_result}.")
                continue
            if update_result["txStatus"] == 1:
                if update_result["txReceipt"]["status"] == 1:
                    if tracked_order.current_state == OrderState.PENDING_CANCEL:
                        if not tracked_order.is_approval_request:
                            order_update: OrderUpdate = OrderUpdate(
                                trading_pair=tracked_order.trading_pair,
                                client_order_id=tracked_order.client_order_id,
                                update_timestamp=self.current_timestamp,
                                new_state=OrderState.CANCELED
                            )
                            self._order_tracker.process_order_update(order_update)

                        elif tracked_order.is_approval_request:
                            order_update: OrderUpdate = OrderUpdate(
                                trading_pair=tracked_order.trading_pair,
                                client_order_id=tracked_order.client_order_id,
                                update_timestamp=self.current_timestamp,
                                new_state=OrderState.CANCELED
                            )
                            token_symbol: str = self.get_token_symbol_from_approval_order_id(
                                tracked_order.client_order_id
                            )
                            self.trigger_event(
                                TokenApprovalEvent.ApprovalCancelled,
                                TokenApprovalCancelledEvent(
                                    self.current_timestamp,
                                    self.connector_name,
                                    token_symbol
                                )
                            )
                            self.logger().info(f"Token approval for {tracked_order.client_order_id} on "
                                               f"{self.connector_name} has been canceled.")
                            self.stop_tracking_order(tracked_order.client_order_id)

    async def update_order_status(self, tracked_orders: List[GatewayInFlightOrder]):
        """
        Calls REST API to get status update for each in-flight amm orders.
        """
        if len(tracked_orders) < 1:
            return

        # split canceled and non-canceled orders
        tx_hash_list: List[str] = await safe_gather(
            *[tracked_order.get_exchange_order_id() for tracked_order in tracked_orders]
        )
        self.logger().info(
            "Polling for order status updates of %d orders. Transaction hashes: %s",
            len(tracked_orders),
            tx_hash_list
        )
        update_results: List[Union[Dict[str, Any], Exception]] = await safe_gather(*[
            self._get_gateway_instance().get_transaction_status(
                self.chain,
                self.network,
                tx_hash
            )
            for tx_hash in tx_hash_list
        ], return_exceptions=True)
        for tracked_order, tx_details in zip(tracked_orders, update_results):
            if isinstance(tx_details, Exception):
                self.logger().error(f"An error occurred fetching transaction status of {tracked_order.client_order_id}")
                continue
            if "txHash" not in tx_details:
                self.logger().error(f"No txHash field for transaction status of {tracked_order.client_order_id}: "
                                    f"{tx_details}.")
                continue
            tx_status: int = tx_details["txStatus"]
            tx_receipt: Optional[Dict[str, Any]] = tx_details["txReceipt"]
            if tx_status == 1 and (tx_receipt is not None and tx_receipt.get("status") == 1):
                gas_used: int = tx_receipt["gasUsed"]
                gas_price: Decimal = tracked_order.gas_price
                fee: Decimal = Decimal(str(gas_used)) * Decimal(str(gas_price)) / Decimal(str(1e9))

                self.process_trade_fill_update(tracked_order=tracked_order, fee=fee)

                order_update: OrderUpdate = OrderUpdate(
                    client_order_id=tracked_order.client_order_id,
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=self.current_timestamp,
                    new_state=OrderState.FILLED,
                )
                self._order_tracker.process_order_update(order_update)
            elif tx_status in [0, 2, 3]:
                # 0: in the mempool but we dont have data to guess its status
                # 2: in the mempool and likely to succeed
                # 3: in the mempool and likely to fail
                pass

            elif tx_status == -1 or (tx_receipt is not None and tx_receipt.get("status") == 0):
                self.logger().network(
                    f"Error fetching transaction status for the order {tracked_order.client_order_id}: {tx_details}.",
                    app_warning_msg=f"Failed to fetch transaction status for the order {tracked_order.client_order_id}."
                )
                await self._order_tracker.process_order_not_found(tracked_order.client_order_id)

    def has_allowances(self) -> bool:
        """
        Checks if all tokens have allowance (an amount approved)
        """
        allowances_available = all(amount > s_decimal_0 for amount in self._allowances.values())
        return ((len(self._allowances.values()) == len(self._tokens)) and
                (allowances_available))

    async def _update_nonce(self, new_nonce: Optional[int] = None):
        """
        Call the gateway API to get the current nonce for self.address
        """
        if not new_nonce:
            resp_json: Dict[str, Any] = await self._get_gateway_instance().get_evm_nonce(self.chain, self.network, self.address)
            new_nonce: int = resp_json.get("nonce")

        self._nonce = new_nonce

    async def _execute_cancel(self, order_id: str, cancel_age: int) -> Optional[str]:
        """
        Cancel an existing order if the age of the order is greater than its cancel_age,
        and if the order is not done or already in the cancelling state.
        """
        try:
            tracked_order: GatewayInFlightOrder = self._order_tracker.fetch_order(client_order_id=order_id)
            if tracked_order is None:
                self.logger().error(f"The order {order_id} is not being tracked.")
                raise ValueError(f"The order {order_id} is not being tracked.")

            if (self.current_timestamp - tracked_order.creation_timestamp) < cancel_age:
                return None

            if tracked_order.is_done:
                return None

            if tracked_order.is_pending_cancel_confirmation:
                return order_id

            self.logger().info(f"The blockchain transaction for {order_id} with nonce {tracked_order.nonce} has "
                               f"expired. Canceling the order...")
            resp: Dict[str, Any] = await self._get_gateway_instance().cancel_evm_transaction(
                self.chain,
                self.network,
                self.address,
                tracked_order.nonce
            )

            tx_hash: Optional[str] = resp.get("txHash")
            if tx_hash is not None:
                tracked_order.cancel_tx_hash = tx_hash
            else:
                raise EnvironmentError(f"Missing txHash from cancel_evm_transaction() response: {resp}.")

            order_update: OrderUpdate = OrderUpdate(
                client_order_id=order_id,
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=OrderState.PENDING_CANCEL
            )
            self._order_tracker.process_order_update(order_update)

            return order_id
        except asyncio.CancelledError:
            raise
        except Exception as err:
            self.logger().error(
                f"Failed to cancel order {order_id}: {str(err)}.",
                exc_info=True
            )

    async def cancel_outdated_orders(self, cancel_age: int) -> List[CancellationResult]:
        """
        Iterate through all known orders and cancel them if their age is greater than cancel_age.
        """
        incomplete_orders: List[GatewayInFlightOrder] = []

        # Incomplete Approval Requests
        incomplete_orders.extend([
            o for o in self.approval_orders
            if o.is_pending_approval
        ])
        # Incomplete Active Orders
        incomplete_orders.extend([
            o for o in self.amm_orders
            if not o.is_done
        ])

        if len(incomplete_orders) < 1:
            return []

        timeout_seconds: float = 30.0
        canceling_id_set: Set[str] = set([o.client_order_id for o in incomplete_orders])
        sent_cancellations: List[CancellationResult] = []

        try:
            async with timeout(timeout_seconds):
                for incomplete_order in incomplete_orders:
                    try:
                        canceling_order_id: Optional[str] = await self._execute_cancel(
                            incomplete_order.client_order_id,
                            cancel_age
                        )
                    except Exception:
                        continue
                    if canceling_order_id is not None:
                        canceling_id_set.remove(canceling_order_id)
                        sent_cancellations.append(CancellationResult(canceling_order_id, True))
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().network(
                "Unexpected error cancelling outdated orders.",
                exc_info=True,
                app_warning_msg=f"Failed to cancel orders on {self.chain}-{self.network}."
            )

        skipped_cancellations: List[CancellationResult] = [CancellationResult(oid, False) for oid in canceling_id_set]
        return sent_cancellations + skipped_cancellations



import argparse
from typing import TYPE_CHECKING, Any, List

from hummingbot.client.command.connect_command import OPTIONS as CONNECT_OPTIONS
from hummingbot.exceptions import ArgumentParserError

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)

    def exit(self, status=0, message=None):
        pass

    def print_help(self, file=None):
        pass

    @property
    def subparser_action(self):
        for action in self._actions:
            if isinstance(action, argparse._SubParsersAction):
                return action

    @property
    def commands(self) -> List[str]:
        return list(self.subparser_action._name_parser_map.keys())

    def subcommands_from(self, top_level_command: str) -> List[str]:
        parser: argparse.ArgumentParser = self.subparser_action._name_parser_map.get(top_level_command)
        if parser is None:
            return []
        subcommands = parser._optionals._option_string_actions.keys()
        filtered = list(filter(lambda sub: sub.startswith("--"), subcommands))
        return filtered


def load_parser(hummingbot: "HummingbotApplication", command_tabs) -> [ThrowingArgumentParser, Any]:
    parser = ThrowingArgumentParser(prog="", add_help=False)
    subparsers = parser.add_subparsers()

    connect_parser = subparsers.add_parser("connect", help="List available exchanges and add API keys to them")
    connect_parser.add_argument("option", nargs="?", choices=CONNECT_OPTIONS, help="Name of the exchange that you want to connect")
    connect_parser.set_defaults(func=hummingbot.connect)

    create_parser = subparsers.add_parser("create", help="Create a new bot")
    create_parser.add_argument("--script-config", dest="script_to_config", nargs="?", default=None, help="Name of the v2 strategy")
    create_parser.add_argument("--controller-config", dest="controller_name", nargs="?", default=None, help="Name of the controller")
    create_parser.set_defaults(func=hummingbot.create)

    import_parser = subparsers.add_parser("import", help="Import an existing bot by loading the configuration file")
    import_parser.add_argument("file_name", nargs="?", default=None, help="Name of the configuration file")
    import_parser.set_defaults(func=hummingbot.import_command)

    help_parser = subparsers.add_parser("help", help="List available commands")
    help_parser.add_argument("command", nargs="?", default="all", help="Enter ")
    help_parser.set_defaults(func=hummingbot.help)

    balance_parser = subparsers.add_parser("balance", help="Display your asset balances across all connected exchanges")
    balance_parser.add_argument("option", nargs="?", choices=["limit", "paper"], default=None,
                                help="Option for balance configuration")
    balance_parser.add_argument("args", nargs="*")
    balance_parser.set_defaults(func=hummingbot.balance)

    config_parser = subparsers.add_parser("config", help="Display the current bot's configuration")
    config_parser.add_argument("key", nargs="?", default=None, help="Name of the parameter you want to change")
    config_parser.add_argument("value", nargs="?", default=None, help="New value for the parameter")
    config_parser.set_defaults(func=hummingbot.config)

    start_parser = subparsers.add_parser("start", help="Start the current bot")
    # start_parser.add_argument("--log-level", help="Level of logging")
    start_parser.add_argument("--script", type=str, dest="script", help="Script strategy file name")
    start_parser.add_argument("--conf", type=str, dest="conf", help="Script config file name")

    start_parser.set_defaults(func=hummingbot.start)

    stop_parser = subparsers.add_parser('stop', help="Stop the current bot")
    stop_parser.set_defaults(func=hummingbot.stop)

    status_parser = subparsers.add_parser("status", help="Get the market status of the current bot")
    status_parser.add_argument("--live", default=False, action="store_true", dest="live", help="Show status updates")
    status_parser.set_defaults(func=hummingbot.status)

    history_parser = subparsers.add_parser("history", help="See the past performance of the current bot")
    history_parser.add_argument("-d", "--days", type=float, default=0, dest="days",
                                help="How many days in the past (can be decimal value)")
    history_parser.add_argument("-v", "--verbose", action="store_true", default=False,
                                dest="verbose", help="List all trades")
    history_parser.add_argument("-p", "--precision", default=None, type=int,
                                dest="precision", help="Level of precions for values displayed")
    history_parser.set_defaults(func=hummingbot.history)

    gateway_parser = subparsers.add_parser("gateway", help="Helper commands for Gateway server.")
    gateway_subparsers = gateway_parser.add_subparsers()

    gateway_balance_parser = gateway_subparsers.add_parser("balance", help="Display your asset balances and allowances across all connected gateway connectors")
    gateway_balance_parser.add_argument("connector_chain_network", nargs="?", default=None, help="Name of connector_chain_network balance and allowance you want to fetch")
    gateway_balance_parser.set_defaults(func=hummingbot.gateway_balance)

    gateway_config_parser = gateway_subparsers.add_parser("config", help="View or update gateway configuration")
    gateway_config_parser.add_argument("key", nargs="?", default=None, help="Name of the parameter you want to view/change")
    gateway_config_parser.add_argument("value", nargs="?", default=None, help="New value for the parameter")
    gateway_config_parser.set_defaults(func=hummingbot.gateway_config)

    gateway_connect_parser = gateway_subparsers.add_parser("connect", help="Create/view connection info for gateway connector")
    gateway_connect_parser.add_argument("connector", nargs="?", default=None, help="Name of connector you want to create a profile for")
    gateway_connect_parser.set_defaults(func=hummingbot.gateway_connect)

    gateway_connector_tokens_parser = gateway_subparsers.add_parser("connector-tokens", help="Report token balances for gateway connectors")
    gateway_connector_tokens_parser.add_argument("connector_chain_network", nargs="?", default=None, help="Name of connector_chain_network you want to edit reported tokens for")
    gateway_connector_tokens_parser.add_argument("new_tokens", nargs="?", default=None, help="Report balance of these tokens - separate multiple tokens with commas (,)")
    gateway_connector_tokens_parser.set_defaults(func=hummingbot.gateway_connector_tokens)

    gateway_approve_tokens_parser = gateway_subparsers.add_parser("approve-tokens", help="Approve tokens for gateway connectors")
    gateway_approve_tokens_parser.add_argument("connector_chain_network", nargs="?", default=None, help="Name of connector you want to approve tokens for")
    gateway_approve_tokens_parser.add_argument("tokens", nargs="?", default=None, help="Approve these tokens")
    gateway_approve_tokens_parser.set_defaults(func=hummingbot.gateway_approve_tokens)

    gateway_cert_parser = gateway_subparsers.add_parser("generate-certs", help="Create ssl certifcate for gateway")
    gateway_cert_parser.set_defaults(func=hummingbot.generate_certs)

    gateway_list_parser = gateway_subparsers.add_parser("list", help="List gateway connectors and chains and tiers")
    gateway_list_parser.set_defaults(func=hummingbot.gateway_list)

    gateway_test_parser = gateway_subparsers.add_parser("test-connection", help="Ping gateway api server")
    gateway_test_parser.set_defaults(func=hummingbot.test_connection)

    exit_parser = subparsers.add_parser("exit", help="Exit and cancel all outstanding orders")
    exit_parser.add_argument("-f", "--force", action="store_true", help="Force exit without canceling outstanding orders",
                             default=False)
    exit_parser.set_defaults(func=hummingbot.exit)

    export_parser = subparsers.add_parser("export", help="Export secure information")
    export_parser.add_argument("option", nargs="?", choices=("keys", "trades"), help="Export choices")
    export_parser.set_defaults(func=hummingbot.export)

    ticker_parser = subparsers.add_parser("ticker", help="Show market ticker of current order book")
    ticker_parser.add_argument("--live", default=False, action="store_true", dest="live", help="Show ticker updates")
    ticker_parser.add_argument("--exchange", type=str, dest="exchange", help="The exchange of the market")
    ticker_parser.add_argument("--market", type=str, dest="market", help="The market (trading pair) of the order book")
    ticker_parser.set_defaults(func=hummingbot.ticker)

    previous_strategy_parser = subparsers.add_parser("previous", help="Imports the last strategy used")
    previous_strategy_parser.add_argument("option", nargs="?", choices=["Yes,No"], default=None)
    previous_strategy_parser.set_defaults(func=hummingbot.previous_strategy)

    mqtt_parser = subparsers.add_parser("mqtt", help="Manage MQTT Bridge to Message brokers")
    mqtt_subparsers = mqtt_parser.add_subparsers()
    mqtt_start_parser = mqtt_subparsers.add_parser("start", help="Start the MQTT Bridge")
    mqtt_start_parser.add_argument(
        "-t",
        "--timeout",
        default=30.0,
        type=float,
        dest="timeout",
        help="Bridge connection timeout"
    )
    mqtt_start_parser.set_defaults(func=hummingbot.mqtt_start)
    mqtt_stop_parser = mqtt_subparsers.add_parser("stop", help="Stop the MQTT Bridge")
    mqtt_stop_parser.set_defaults(func=hummingbot.mqtt_stop)
    mqtt_restart_parser = mqtt_subparsers.add_parser("restart", help="Restart the MQTT Bridge")
    mqtt_restart_parser.add_argument(
        "-t",
        "--timeout",
        default=30.0,
        type=float,
        dest="timeout",
        help="Bridge connection timeout"
    )
    mqtt_restart_parser.set_defaults(func=hummingbot.mqtt_restart)

    # add shortcuts so they appear in command help
    shortcuts = hummingbot.client_config_map.command_shortcuts
    for shortcut in shortcuts:
        help_str = shortcut.help
        command = shortcut.command
        shortcut_parser = subparsers.add_parser(command, help=help_str)
        args = shortcut.arguments
        for i in range(len(args)):
            shortcut_parser.add_argument(f'${i + 1}', help=args[i])

    rate_parser = subparsers.add_parser('rate', help="Show rate of a given trading pair")
    rate_parser.add_argument("-p", "--pair", default=None,
                             dest="pair", help="The market trading pair for which you want to get a rate.")
    rate_parser.add_argument("-t", "--token", default=None,
                             dest="token", help="The token who's value you want to get.")
    rate_parser.set_defaults(func=hummingbot.rate)

    for name, command_tab in command_tabs.items():
        o_parser = subparsers.add_parser(name, help=command_tab.tab_class.get_command_help_message())
        for arg_name, arg_properties in command_tab.tab_class.get_command_arguments().items():
            o_parser.add_argument(arg_name, **arg_properties)
        o_parser.add_argument("-c", "--close", default=False, action="store_true", dest="close",
                              help=f"To close the {name} tab.")

    return parser



import importlib
import inspect
import os
from decimal import Decimal
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
import yaml

from hummingbot.client import settings
from hummingbot.core.data_type.common import TradeType
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.exceptions import InvalidController
from hummingbot.strategy_v2.backtesting.backtesting_data_provider import BacktestingDataProvider
from hummingbot.strategy_v2.backtesting.executor_simulator_base import ExecutorSimulation
from hummingbot.strategy_v2.backtesting.executors_simulator.dca_executor_simulator import DCAExecutorSimulator
from hummingbot.strategy_v2.backtesting.executors_simulator.position_executor_simulator import PositionExecutorSimulator
from hummingbot.strategy_v2.controllers.controller_base import ControllerConfigBase
from hummingbot.strategy_v2.controllers.directional_trading_controller_base import (
    DirectionalTradingControllerConfigBase,
)
from hummingbot.strategy_v2.controllers.market_making_controller_base import MarketMakingControllerConfigBase
from hummingbot.strategy_v2.executors.dca_executor.data_types import DCAExecutorConfig
from hummingbot.strategy_v2.executors.position_executor.data_types import PositionExecutorConfig
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executor_actions import CreateExecutorAction, StopExecutorAction
from hummingbot.strategy_v2.models.executors import CloseType
from hummingbot.strategy_v2.models.executors_info import ExecutorInfo


class BacktestingEngineBase:
    def __init__(self):
        self.controller = None
        self.backtesting_resolution = None
        self.backtesting_data_provider = BacktestingDataProvider(connectors={})
        self.position_executor_simulator = PositionExecutorSimulator()
        self.dca_executor_simulator = DCAExecutorSimulator()

    @classmethod
    def load_controller_config(cls,
                               config_path: str,
                               controllers_conf_dir_path: str = settings.CONTROLLERS_CONF_DIR_PATH) -> Dict:
        full_path = os.path.join(controllers_conf_dir_path, config_path)
        with open(full_path, 'r') as file:
            config_data = yaml.safe_load(file)
        return config_data

    @classmethod
    def get_controller_config_instance_from_yml(cls,
                                                config_path: str,
                                                controllers_conf_dir_path: str = settings.CONTROLLERS_CONF_DIR_PATH,
                                                controllers_module: str = settings.CONTROLLERS_MODULE) -> ControllerConfigBase:
        config_data = cls.load_controller_config(config_path, controllers_conf_dir_path)
        return cls.get_controller_config_instance_from_dict(config_data, controllers_module)

    @classmethod
    def get_controller_config_instance_from_dict(cls,
                                                 config_data: dict,
                                                 controllers_module: str = settings.CONTROLLERS_MODULE) -> ControllerConfigBase:
        controller_type = config_data.get('controller_type')
        controller_name = config_data.get('controller_name')

        if not controller_type or not controller_name:
            raise ValueError("Missing controller_type or controller_name in the configuration.")

        module_path = f"{controllers_module}.{controller_type}.{controller_name}"
        module = importlib.import_module(module_path)

        config_class = next((member for member_name, member in inspect.getmembers(module)
                             if inspect.isclass(member) and member not in [ControllerConfigBase,
                                                                           MarketMakingControllerConfigBase,
                                                                           DirectionalTradingControllerConfigBase]
                             and (issubclass(member, ControllerConfigBase))), None)
        if not config_class:
            raise InvalidController(f"No configuration class found in the module {controller_name}.")

        return config_class(**config_data)

    async def run_backtesting(self,
                              controller_config: ControllerConfigBase,
                              start: int, end: int,
                              backtesting_resolution: str = "1m",
                              trade_cost=0.0006):
        # Load historical candles
        controller_class = controller_config.get_controller_class()
        self.backtesting_data_provider.update_backtesting_time(start, end)
        await self.backtesting_data_provider.initialize_trading_rules(controller_config.connector_name)
        self.controller = controller_class(config=controller_config, market_data_provider=self.backtesting_data_provider,
                                           actions_queue=None)
        self.backtesting_resolution = backtesting_resolution
        await self.initialize_backtesting_data_provider()
        await self.controller.update_processed_data()
        executors_info = await self.simulate_execution(trade_cost=trade_cost)
        results = self.summarize_results(executors_info, controller_config.total_amount_quote)
        return {
            "executors": executors_info,
            "results": results,
            "processed_data": self.controller.processed_data,
        }

    async def initialize_backtesting_data_provider(self):
        backtesting_config = CandlesConfig(
            connector=self.controller.config.connector_name,
            trading_pair=self.controller.config.trading_pair,
            interval=self.backtesting_resolution
        )
        await self.controller.market_data_provider.initialize_candles_feed(backtesting_config)
        for config in self.controller.config.candles_config:
            await self.controller.market_data_provider.initialize_candles_feed(config)

    async def simulate_execution(self, trade_cost: float) -> list:
        """
        Simulates market making strategy over historical data, considering trading costs.

        Args:
            trade_cost (float): The cost per trade.

        Returns:
            List[ExecutorInfo]: List of executor information objects detailing the simulation results.
        """
        processed_features = self.prepare_market_data()
        self.active_executor_simulations: List[ExecutorSimulation] = []
        self.stopped_executors_info: List[ExecutorInfo] = []
        for i, row in processed_features.iterrows():
            await self.update_state(row)
            for action in self.controller.determine_executor_actions():
                if isinstance(action, CreateExecutorAction):
                    executor_simulation = self.simulate_executor(action.executor_config, processed_features.loc[i:], trade_cost)
                    if executor_simulation.close_type != CloseType.FAILED:
                        self.manage_active_executors(executor_simulation)
                elif isinstance(action, StopExecutorAction):
                    self.handle_stop_action(action, row["timestamp"])

        return self.controller.executors_info

    async def update_state(self, row):
        key = f"{self.controller.config.connector_name}_{self.controller.config.trading_pair}"
        self.controller.market_data_provider.prices = {key: Decimal(row["close_bt"])}
        self.controller.market_data_provider._time = row["timestamp"]
        self.controller.processed_data.update(row.to_dict())
        self.update_executors_info(row["timestamp"])

    def update_executors_info(self, timestamp: float):
        active_executors_info = []
        simulations_to_remove = []
        for executor in self.active_executor_simulations:
            executor_info = executor.get_executor_info_at_timestamp(timestamp)
            if executor_info.status == RunnableStatus.TERMINATED:
                self.stopped_executors_info.append(executor_info)
                simulations_to_remove.append(executor.config.id)
            else:
                active_executors_info.append(executor_info)
        self.active_executor_simulations = [es for es in self.active_executor_simulations if es.config.id not in simulations_to_remove]
        self.controller.executors_info = active_executors_info + self.stopped_executors_info

    async def update_processed_data(self, row: pd.Series):
        """
        Updates processed data in the controller with the current price and timestamp.

        Args:
            row (pd.Series): The current row of market data.
        """
        raise NotImplementedError("update_processed_data method must be implemented in a subclass.")

    def prepare_market_data(self) -> pd.DataFrame:
        """
        Prepares market data by merging candle data with strategy features, filling missing values.

        Returns:
            pd.DataFrame: The prepared market data with necessary features.
        """
        backtesting_candles = self.controller.market_data_provider.get_candles_df(
            connector_name=self.controller.config.connector_name,
            trading_pair=self.controller.config.trading_pair,
            interval=self.backtesting_resolution
        ).add_suffix("_bt")

        if "features" not in self.controller.processed_data:
            backtesting_candles["reference_price"] = backtesting_candles["close_bt"]
            backtesting_candles["spread_multiplier"] = 1
            backtesting_candles["signal"] = 0
        else:
            backtesting_candles = pd.merge_asof(backtesting_candles, self.controller.processed_data["features"],
                                                left_on="timestamp_bt", right_on="timestamp",
                                                direction="backward")
        backtesting_candles["timestamp"] = backtesting_candles["timestamp_bt"]
        backtesting_candles["open"] = backtesting_candles["open_bt"]
        backtesting_candles["high"] = backtesting_candles["high_bt"]
        backtesting_candles["low"] = backtesting_candles["low_bt"]
        backtesting_candles["close"] = backtesting_candles["close_bt"]
        backtesting_candles["volume"] = backtesting_candles["volume_bt"]
        backtesting_candles.dropna(inplace=True)
        self.controller.processed_data["features"] = backtesting_candles
        return backtesting_candles

    def simulate_executor(self, config: Union[PositionExecutorConfig, DCAExecutorConfig], df: pd.DataFrame,
                          trade_cost: float) -> Optional[ExecutorSimulation]:
        """
        Simulates the execution of a trading strategy given a configuration.

        Args:
            config (PositionExecutorConfig): The configuration of the executor.
            df (pd.DataFrame): DataFrame containing the market data from the start time.
            trade_cost (float): The cost per trade.

        Returns:
            ExecutorSimulation: The results of the simulation.
        """
        if isinstance(config, DCAExecutorConfig):
            return self.dca_executor_simulator.simulate(df, config, trade_cost)
        elif isinstance(config, PositionExecutorConfig):
            return self.position_executor_simulator.simulate(df, config, trade_cost)
        return None

    def manage_active_executors(self, simulation: ExecutorSimulation):
        """
        Manages the list of active executors based on the simulation results.

        Args:
            simulation (ExecutorSimulation): The simulation results of the current executor.
            active_executors (list): The list of active executors.
        """
        if not simulation.executor_simulation.empty:
            self.active_executor_simulations.append(simulation)

    def handle_stop_action(self, action: StopExecutorAction, timestamp: pd.Timestamp):
        """
        Handles stop actions for executors, terminating them as required.

        Args:
            action (StopExecutorAction): The action indicating which executor to stop.
            active_executors (list): The list of active executors.
            timestamp (pd.Timestamp): The current timestamp.
        """
        for executor in self.active_executor_simulations:
            executor_info = executor.get_executor_info_at_timestamp(timestamp)
            if executor_info.config.id == action.executor_id:
                executor_info.status = RunnableStatus.TERMINATED
                executor_info.close_type = CloseType.EARLY_STOP
                executor_info.is_active = False
                executor_info.close_timestamp = timestamp
                self.stopped_executors_info.append(executor_info)
                self.active_executor_simulations.remove(executor)

    @staticmethod
    def summarize_results(executors_info: List, total_amount_quote: float = 1000):
        if len(executors_info) > 0:
            executors_df = pd.DataFrame([ei.to_dict() for ei in executors_info])
            net_pnl_quote = executors_df["net_pnl_quote"].sum()
            total_executors = executors_df.shape[0]
            executors_with_position = executors_df[executors_df["net_pnl_quote"] != 0]
            total_executors_with_position = executors_with_position.shape[0]
            total_volume = executors_with_position["filled_amount_quote"].sum()
            total_long = (executors_with_position["side"] == TradeType.BUY).sum()
            total_short = (executors_with_position["side"] == TradeType.SELL).sum()
            correct_long = ((executors_with_position["side"] == TradeType.BUY) & (executors_with_position["net_pnl_quote"] > 0)).sum()
            correct_short = ((executors_with_position["side"] == TradeType.SELL) & (executors_with_position["net_pnl_quote"] > 0)).sum()
            accuracy_long = correct_long / total_long if total_long > 0 else 0
            accuracy_short = correct_short / total_short if total_short > 0 else 0
            executors_df["close_type_name"] = executors_df["close_type"].apply(lambda x: x.name)
            close_types = executors_df.groupby("close_type_name")["timestamp"].count().to_dict()
            executors_with_position = executors_df[executors_df["net_pnl_quote"] != 0].copy()
            # Additional metrics
            total_positions = executors_with_position.shape[0]
            win_signals = executors_with_position[executors_with_position["net_pnl_quote"] > 0]
            loss_signals = executors_with_position[executors_with_position["net_pnl_quote"] < 0]
            accuracy = win_signals.shape[0] / total_positions
            cumulative_returns = executors_with_position["net_pnl_quote"].cumsum()
            executors_with_position["cumulative_returns"] = cumulative_returns
            executors_with_position["cumulative_volume"] = executors_with_position["filled_amount_quote"].cumsum()
            executors_with_position["inventory"] = total_amount_quote + cumulative_returns

            peak = np.maximum.accumulate(cumulative_returns)
            drawdown = (cumulative_returns - peak)
            max_draw_down = np.min(drawdown)
            max_drawdown_pct = max_draw_down / executors_with_position["inventory"].iloc[0]
            returns = pd.to_numeric(
                executors_with_position["cumulative_returns"] / executors_with_position["cumulative_volume"])
            sharpe_ratio = returns.mean() / returns.std() if len(returns) > 1 else 0
            total_won = win_signals.loc[:, "net_pnl_quote"].sum()
            total_loss = - loss_signals.loc[:, "net_pnl_quote"].sum()
            profit_factor = total_won / total_loss if total_loss > 0 else 1
            net_pnl_pct = net_pnl_quote / total_amount_quote

            return {
                "net_pnl": float(net_pnl_pct),
                "net_pnl_quote": float(net_pnl_quote),
                "total_executors": int(total_executors),
                "total_executors_with_position": int(total_executors_with_position),
                "total_volume": float(total_volume),
                "total_long": int(total_long),
                "total_short": int(total_short),
                "close_types": close_types,
                "accuracy_long": float(accuracy_long),
                "accuracy_short": float(accuracy_short),
                "total_positions": int(total_positions),
                "accuracy": float(accuracy),
                "max_drawdown_usd": float(max_draw_down),
                "max_drawdown_pct": float(max_drawdown_pct),
                "sharpe_ratio": float(sharpe_ratio),
                "profit_factor": float(profit_factor),
                "win_signals": int(win_signals.shape[0]),
                "loss_signals": int(loss_signals.shape[0]),
            }
        return {
            "net_pnl": 0,
            "net_pnl_quote": 0,
            "total_executors": 0,
            "total_executors_with_position": 0,
            "total_volume": 0,
            "total_long": 0,
            "total_short": 0,
            "close_types": 0,
            "accuracy_long": 0,
            "accuracy_short": 0,
            "total_positions": 0,
            "accuracy": 0,
            "max_drawdown_usd": 0,
            "max_drawdown_pct": 0,
            "sharpe_ratio": 0,
            "profit_factor": 0,
            "win_signals": 0,
            "loss_signals": 0,
        }



import asyncio
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from hummingbot.core.data_type.funding_info import FundingInfo
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource


class PerpetualAPIOrderBookDataSource(OrderBookTrackerDataSource, ABC):
    def __init__(self, trading_pairs: List[str]):
        super().__init__(trading_pairs)
        self._funding_info_messages_queue_key = "funding_info"

    @abstractmethod
    async def get_funding_info(self, trading_pair: str) -> FundingInfo:
        """
        Return the funding information for a single trading pair.
        """
        raise NotImplementedError

    async def listen_for_funding_info(self, output: asyncio.Queue):
        """
        Reads the funding info events queue and updates the local funding info information.
        """
        message_queue = self._message_queue[self._funding_info_messages_queue_key]
        while True:
            try:
                funding_info_event = await message_queue.get()
                await self._parse_funding_info_message(raw_message=funding_info_event, message_queue=output)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error when processing public funding info updates from exchange")

    @abstractmethod
    async def _parse_funding_info_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        raise NotImplementedError

    def _get_messages_queue_keys(self) -> List[str]:
        return [
            self._snapshot_messages_queue_key,
            self._diff_messages_queue_key,
            self._trade_messages_queue_key,
            self._funding_info_messages_queue_key,
        ]



import asyncio
import concurrent.futures
import threading
import unittest
from test.isolated_asyncio_wrapper_test_case import IsolatedAsyncioWrapperTestCase, async_to_sync


class TestIsolatedAsyncioWrapperTestCase(unittest.IsolatedAsyncioTestCase):
    def test_setUpClass_with_existing_loop(self):
        self.main_loop = asyncio.get_event_loop()

        IsolatedAsyncioWrapperTestCase.setUpClass()
        self.assertIsNotNone(IsolatedAsyncioWrapperTestCase.main_event_loop)
        self.assertEqual(self.main_loop, IsolatedAsyncioWrapperTestCase.main_event_loop)

        self.main_loop = None

    def test_setUpClass_with_new_loop(self):
        self.main_loop = asyncio.get_event_loop()
        self.local_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.local_loop)

        IsolatedAsyncioWrapperTestCase.setUpClass()
        self.assertIsNotNone(IsolatedAsyncioWrapperTestCase.main_event_loop)
        self.assertEqual(self.local_loop, IsolatedAsyncioWrapperTestCase.main_event_loop)

        self.local_loop.close()
        asyncio.set_event_loop(self.main_loop)
        self.main_loop = None

    def test_setUpClass_without_existing_loop(self):
        def run_test_in_thread(future):
            asyncio.set_event_loop(None)

            try:
                IsolatedAsyncioWrapperTestCase.setUpClass()
                self.assertIsNotNone(IsolatedAsyncioWrapperTestCase.main_event_loop)
                self.assertEqual(asyncio.get_event_loop(), IsolatedAsyncioWrapperTestCase.main_event_loop)
            except Exception as e:
                future.set_exception(e)
            else:
                future.set_result(None)

        future = concurrent.futures.Future()
        thread = threading.Thread(target=run_test_in_thread, args=(future,))
        thread.start()
        thread.join()
        future.result()

    def test_tearDownClass_with_existing_loop(self):
        self.main_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.main_loop)

        IsolatedAsyncioWrapperTestCase.main_event_loop = self.main_loop
        IsolatedAsyncioWrapperTestCase.tearDownClass()
        self.assertEqual(self.main_loop, asyncio.get_event_loop())

        self.main_loop.close()
        asyncio.set_event_loop(None)
        self.main_loop = None

    def test_tearDownClass_without_existing_loop(self):
        # Close the main event loop if it exists
        def run_test_in_thread(future):
            try:
                asyncio.set_event_loop(None)
                asyncio.get_event_loop()
            except RuntimeError:
                pass

            try:
                IsolatedAsyncioWrapperTestCase.tearDownClass()
                asyncio.get_event_loop()
            except Exception as e:
                future.set_exception(e)
            else:
                future.set_result(None)

        future = concurrent.futures.Future()
        thread = threading.Thread(target=run_test_in_thread, args=(future,))
        thread.start()
        thread.join()
        future.result()

    async def _dummy_coro(self, name, delay):
        """A dummy coroutine that simply sleeps for a delay."""
        await asyncio.sleep(delay)

    async def _dummy_coro_to_await(self, name, delay):
        """A dummy coroutine that simply sleeps for a delay."""
        await asyncio.sleep(delay)

    async def test_await_task_completion(self):
        # Create some tasks with different coroutine names
        task1 = asyncio.create_task(self._dummy_coro("task1", 0.5))
        task2 = asyncio.create_task(self._dummy_coro("task2", 0.75))
        task3 = asyncio.create_task(self._dummy_coro_to_await("task3", 2))

        # Use the await_task_completion method to wait for task1 and task2 to complete
        self.assertFalse(task1.done())
        self.assertFalse(task2.done())
        self.assertFalse(task3.done())

        await IsolatedAsyncioWrapperTestCase.await_task_completion(["_dummy_coro"])

        # At this point, task1 and task2 should be done, but task3 should still be running
        self.assertTrue(task1.done())
        self.assertTrue(task2.done())
        self.assertFalse(task3.done())

        # Now wait for task3 to complete as well
        await IsolatedAsyncioWrapperTestCase.await_task_completion("_dummy_coro_to_await")
        self.assertTrue(task3.done())


class TestAsyncToSyncInLoop(unittest.TestCase):
    @async_to_sync
    async def async_add(self, a: int, b: int) -> int:
        await asyncio.sleep(0.1)
        return a + b

    def test_async_add(self):
        result = self.async_add(1, 2)
        self.assertEqual(result, 3)

    @async_to_sync
    async def async_raise_exception(self) -> None:
        await asyncio.sleep(0.1)
        raise ValueError("Test exception")

    def test_async_raise_exception(self):
        with self.assertRaises(ValueError) as context:
            self.async_raise_exception()
        self.assertEqual(str(context.exception), "Test exception")

    def test_main_event_loop_unchanged(self):
        # Save the current event loop
        try:
            main_loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            main_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(main_loop)

        # Run a function decorated with @async_to_sync
        self.async_add(1, 2)

        # Check that the current event loop is still the same
        self.assertEqual(main_loop, asyncio.get_event_loop())

    def test_main_event_loop_unchanged_after_exception(self):
        # Save the current event loop
        try:
            main_loop = asyncio.get_event_loop()
        except RuntimeError:
            # If no event loop exists, create one
            main_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(main_loop)

        # Run a function decorated with @async_to_sync that raises an exception
        with self.assertRaises(ValueError):
            self.async_raise_exception()

        # Check that the current event loop is still the same
        self.assertEqual(main_loop, asyncio.get_event_loop())


if __name__ == "__main__":
    unittest.main()



import asyncio

from dataclasses import dataclass
from prompt_toolkit.widgets import Button
from typing import Type, Optional

from hummingbot.client.ui.custom_widgets import CustomTextArea
from .tab_base import TabBase


@dataclass
class CommandTab:
    """
    Defines all data points for a tab.
    """
    name: str  # Command name of the tab
    button: Optional[Button]  # Tab toggle button
    close_button: Optional[Button]  # Tab close button
    output_field: Optional[CustomTextArea]  # Output pane where tab messages display
    tab_class: Type[TabBase]  # The tab class (Subclass of TabBase)
    is_selected: bool = False  # If the tab is currently selected by a user
    tab_index: int = 0  # The index position of the tab in relation of all other displayed tabs
    task: Optional[asyncio.Task] = None  # The currently running task, None if there isn't one



import asyncio
import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import List, Optional

from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.event.events import OrderType, TradeType
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple

s_decimal_nan = Decimal("NaN")
s_decimal_0 = Decimal("0")
arbprop_logger: Optional[HummingbotLogger] = None


@dataclass
class ArbProposalSide:
    """
    An arbitrage proposal side which contains info needed for order submission.
    """
    market_info: MarketTradingPairTuple
    is_buy: bool
    quote_price: Decimal
    order_price: Decimal
    amount: Decimal
    extra_flat_fees: List[TokenAmount]
    completed_event: asyncio.Event = asyncio.Event()
    failed_event: asyncio.Event = asyncio.Event()

    def __repr__(self):
        side = "buy" if self.is_buy else "sell"
        return f"Connector: {self.market_info.market.display_name}  Side: {side}  Quote Price: {self.quote_price}  " \
               f"Order Price: {self.order_price}  Amount: {self.amount}  Extra Fees: {self.extra_flat_fees}"

    @property
    def is_completed(self) -> bool:
        return self.completed_event.is_set()

    @property
    def is_failed(self) -> bool:
        return self.failed_event.is_set()

    def set_completed(self):
        self.completed_event.set()

    def set_failed(self):
        self.failed_event.set()


class ArbProposal:
    @classmethod
    def logger(cls) -> HummingbotLogger:
        global arbprop_logger
        if arbprop_logger is None:
            arbprop_logger = logging.getLogger(__name__)
        return arbprop_logger

    """
    An arbitrage proposal which contains 2 sides of the proposal - one buy and one sell.
    """
    def __init__(self, first_side: ArbProposalSide, second_side: ArbProposalSide):
        if first_side.is_buy == second_side.is_buy:
            raise Exception("first_side and second_side must be on different side of buy and sell.")
        self.first_side: ArbProposalSide = first_side
        self.second_side: ArbProposalSide = second_side

    @property
    def has_failed_orders(self) -> bool:
        return any([self.first_side.is_failed, self.second_side.is_failed])

    def profit_pct(
            self,
            rate_source: Optional[RateOracle] = None,
            account_for_fee: bool = False,
    ) -> Decimal:
        """
        Returns a profit in percentage value (e.g. 0.01 for 1% profitability)
        Assumes the base token is the same in both arbitrage sides
        """
        if not rate_source:
            rate_source = RateOracle.get_instance()

        buy_side: ArbProposalSide = self.first_side if self.first_side.is_buy else self.second_side
        sell_side: ArbProposalSide = self.first_side if not self.first_side.is_buy else self.second_side
        base_conversion_pair: str = f"{sell_side.market_info.base_asset}-{buy_side.market_info.base_asset}"
        quote_conversion_pair: str = f"{sell_side.market_info.quote_asset}-{buy_side.market_info.quote_asset}"

        sell_base_to_buy_base_rate: Decimal = Decimal(1)
        sell_quote_to_buy_quote_rate: Decimal = rate_source.get_pair_rate(quote_conversion_pair)

        buy_fee_amount: Decimal = s_decimal_0
        sell_fee_amount: Decimal = s_decimal_0
        result: Decimal = s_decimal_0

        if sell_quote_to_buy_quote_rate and sell_base_to_buy_base_rate:
            if account_for_fee:
                buy_trade_fee: TradeFeeBase = build_trade_fee(
                    exchange=buy_side.market_info.market.name,
                    is_maker=False,
                    base_currency=buy_side.market_info.base_asset,
                    quote_currency=buy_side.market_info.quote_asset,
                    order_type=OrderType.MARKET,
                    order_side=TradeType.BUY,
                    amount=buy_side.amount,
                    price=buy_side.order_price,
                    extra_flat_fees=buy_side.extra_flat_fees
                )
                sell_trade_fee: TradeFeeBase = build_trade_fee(
                    exchange=sell_side.market_info.market.name,
                    is_maker=False,
                    base_currency=sell_side.market_info.base_asset,
                    quote_currency=sell_side.market_info.quote_asset,
                    order_type=OrderType.MARKET,
                    order_side=TradeType.SELL,
                    amount=sell_side.amount,
                    price=sell_side.order_price,
                    extra_flat_fees=sell_side.extra_flat_fees
                )
                buy_fee_amount: Decimal = buy_trade_fee.fee_amount_in_token(
                    trading_pair=buy_side.market_info.trading_pair,
                    price=buy_side.quote_price,
                    order_amount=buy_side.amount,
                    token=buy_side.market_info.quote_asset,
                    rate_source=rate_source
                )
                sell_fee_amount: Decimal = sell_trade_fee.fee_amount_in_token(
                    trading_pair=sell_side.market_info.trading_pair,
                    price=sell_side.quote_price,
                    order_amount=sell_side.amount,
                    token=sell_side.market_info.quote_asset,
                    rate_source=rate_source
                )

            buy_spent_net: Decimal = (buy_side.amount * buy_side.quote_price) + buy_fee_amount
            sell_gained_net: Decimal = (sell_side.amount * sell_side.quote_price) - sell_fee_amount
            sell_gained_net_in_buy_quote_currency: Decimal = (
                sell_gained_net * sell_quote_to_buy_quote_rate / sell_base_to_buy_base_rate
            )

            result: Decimal = (
                ((sell_gained_net_in_buy_quote_currency - buy_spent_net) / buy_spent_net)
                if buy_spent_net != s_decimal_0
                else s_decimal_0
            )
        else:
            self.logger().warning("The arbitrage proposal profitability could not be calculated due to a missing rate"
                                  f" ({base_conversion_pair}={sell_base_to_buy_base_rate},"
                                  f" {quote_conversion_pair}={sell_quote_to_buy_quote_rate})")
        return result

    def __repr__(self):
        return f"First Side - {self.first_side}\nSecond Side - {self.second_side}"

    def copy(self):
        return ArbProposal(
            ArbProposalSide(self.first_side.market_info, self.first_side.is_buy,
                            self.first_side.quote_price, self.first_side.order_price,
                            self.first_side.amount, self.first_side.extra_flat_fees),
            ArbProposalSide(self.second_side.market_info, self.second_side.is_buy,
                            self.second_side.quote_price, self.second_side.order_price,
                            self.second_side.amount, self.second_side.extra_flat_fees)
        )

    async def wait(self):
        return await safe_gather(*[self.first_side.completed_event.wait(), self.second_side.completed_event.wait()])



import hashlib
import hmac
import json
import re
from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any, Callable, Dict, List, Optional, Pattern, Tuple, Union

from aioresponses.core import RequestCall, aioresponses

from hummingbot.connector.test_support.exchange_connector_test import AbstractExchangeConnectorTests
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utilities.oms_connector import oms_connector_constants as CONSTANTS
from hummingbot.connector.utilities.oms_connector.oms_connector_auth import OMSConnectorAuth
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase


class OMSExchangeTests:
    class ExchangeTests(AbstractExchangeConnectorTests.ExchangeConnectorTests, ABC):
        @classmethod
        def setUpClass(cls) -> None:
            super().setUpClass()
            cls.api_key = "someApiKey"
            cls.secret = "someSecret"
            cls.user_id = 20
            cls.time_mock = 1655283229.419752
            cls.nonce = str(int(cls.time_mock * 1e3))
            auth_concat = f"{cls.nonce}{cls.user_id}{cls.api_key}"
            cls.signature = hmac.new(
                key=cls.secret.encode("utf-8"),
                msg=auth_concat.encode("utf-8"),
                digestmod=hashlib.sha256,
            ).hexdigest()
            cls.user_name = "someUserName"
            cls.oms_id = 1
            cls.account_id = 3
            cls.pair_id = 1
            cls.base_asset = "COINALPHA"
            cls.base_id = 26
            cls.quote_asset = "HBOT"
            cls.quote_id = 2
            cls.trading_pair = f"{cls.base_asset}-{cls.quote_asset}"
            cls.exchange_trading_pair = f"{cls.base_asset}{cls.quote_asset}"

        def setUp(self) -> None:
            super().setUp()
            self.exchange._token_id_map[self.quote_id] = self.quote_asset
            self.exchange._token_id_map[self.base_id] = self.base_asset

        @property
        @abstractmethod
        def url_creator(self):
            raise NotImplementedError

        @property
        def all_symbols_url(self):
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_PRODUCTS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?") + ".*")
            return regex_url

        @property
        def latest_prices_url(self) -> Pattern[str]:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_GET_L1_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            return regex_url

        @property
        def network_status_url(self) -> Pattern[str]:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_PING_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            return regex_url

        @property
        def trading_rules_url(self) -> Pattern[str]:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_PRODUCTS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            return regex_url

        @property
        def order_creation_url(self) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_CREATION_ENDPOINT)
            return url

        @property
        def balance_url(self) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ACC_POSITIONS_ENDPOINT)
            return url

        @property
        def all_symbols_request_mock_response(self) -> List[Dict[str, Any]]:
            return self.get_products_resp()

        @property
        def all_symbols_including_invalid_pair_mock_response(self) -> Tuple[str, List[Dict[str, Any]]]:
            resp = self.get_products_resp()
            resp[0]["IsDisable"] = True
            return self.trading_pair, resp

        @property
        def latest_prices_request_mock_response(self) -> Dict[str, Union[int, float]]:
            return {
                "AskOrderCt": 0,
                "AskQty": 1,
                "BestBid": 0.382882,
                "BestOffer": 0.38345,
                "BidOrderCt": 0,
                "BidQty": 0,
                "CurrentDayNotional": 0,
                "CurrentDayNumTrades": 0,
                "CurrentDayPxChange": 0,
                "CurrentDayVolume": 0,
                "InstrumentId": self.pair_id,
                "LastTradeTime": 0,
                "LastTradedPx": float(self.expected_latest_price),
                "LastTradedQty": 0,
                "OMSId": self.oms_id,
                "Rolling24HrNotional": 0,
                "Rolling24HrPxChange": 0,
                "Rolling24HrPxChangePercent": 0,
                "Rolling24HrVolume": 0,
                "Rolling24NumTrades": 0,
                "SessionClose": 0,
                "SessionHigh": 0,
                "SessionLow": 0,
                "SessionOpen": 0,
                "TimeStamp": "0",
                "Volume": 0,
            }

        @property
        def network_status_request_successful_mock_response(self) -> Dict[str, str]:
            return {"msg": "PONG"}

        @property
        def trading_rules_request_mock_response(self) -> List[Dict[str, Any]]:
            return self.get_products_resp()

        @property
        def trading_rules_request_erroneous_mock_response(self):
            resp = self.get_products_resp()
            resp[0].pop("MinimumQuantity")
            return resp

        def get_auth_success_response(self) -> Dict[str, Any]:
            auth_resp = {
                "Authenticated": True,
                "SessionToken": "0e8bbcbc-6ada-482a-a9b4-5d9218ada3f9",
                "User": {
                    "UserId": self.user_id,
                    "UserName": self.user_name,
                    "Email": "",
                    "EmailVerified": True,
                    "AccountId": self.account_id,
                    "OMSId": self.oms_id,
                    "Use2FA": False,
                },
                "Locked": False,
                "Requires2FA": False,
                "EnforceEnable2FA": False,
                "TwoFAType": None,
                "TwoFAToken": None,
                "errormsg": None,
            }
            return auth_resp

        @staticmethod
        def get_auth_failure_response() -> Dict[str, Any]:
            auth_resp = {
                "Authenticated": False,
                "EnforceEnable2FA": False,
                "Locked": False,
                "Requires2FA": False,
                "SessionToken": None,
                "TwoFAToken": None,
                "TwoFAType": None,
                "User": {
                    "AccountId": 0,
                    "Email": None,
                    "EmailVerified": False,
                    "OMSId": 0,
                    "Use2FA": False,
                    "UserId": 0,
                    "UserName": None,
                },
                "errormsg": "User api key not found",
            }
            return auth_resp

        def get_products_resp(self) -> List[Dict[str, Any]]:
            return [
                {
                    "AllowOnlyMarketMakerCounterParty": False,
                    "CreateWithMarketRunning": True,
                    "InstrumentId": self.pair_id,
                    "InstrumentType": "Standard",
                    "IsDisable": False,
                    "MasterDataId": 0,
                    "MinimumPrice": 0.9,
                    "MinimumQuantity": 0.01,
                    "OMSId": self.oms_id,
                    "OtcConvertSizeEnabled": False,
                    "OtcConvertSizeThreshold": 0.0,
                    "OtcTradesPublic": True,
                    "PreviousSessionStatus": "Stopped",
                    "PriceCeilingLimit": 0.0,
                    "PriceCeilingLimitEnabled": False,
                    "PriceCollarConvertToOtcAccountId": 0,
                    "PriceCollarConvertToOtcClientUserId": 0,
                    "PriceCollarConvertToOtcEnabled": False,
                    "PriceCollarConvertToOtcThreshold": 0.0,
                    "PriceCollarEnabled": False,
                    "PriceCollarIndexDifference": 0.0,
                    "PriceCollarPercent": 0.0,
                    "PriceCollarThreshold": 0.0,
                    "PriceFloorLimit": 0.0,
                    "PriceFloorLimitEnabled": False,
                    "PriceIncrement": 1e-05,
                    "PriceTier": 0,
                    "Product1": self.base_id,
                    "Product1Symbol": self.base_asset,
                    "Product2": self.quote_id,
                    "Product2Symbol": self.quote_asset,
                    "QuantityIncrement": 0.0001,
                    "SelfTradePrevention": True,
                    "SessionStatus": "Running",
                    "SessionStatusDateTime": "2022-05-23T17:11:16.422Z",
                    "SortIndex": 0,
                    "Symbol": self.exchange_trading_pair,
                    "VenueId": 1,
                    "VenueInstrumentId": 11,
                    "VenueSymbol": self.exchange_trading_pair,
                }
            ]

        @property
        def order_creation_request_successful_mock_response(self) -> Dict[str, Union[str, int]]:
            return {
                "status": "Accepted",
                "errormsg": "",
                "OrderId": self.expected_exchange_order_id,
            }

        @property
        def balance_request_mock_response_for_base_and_quote(self) -> List[Dict[str, Union[str, int]]]:
            return [
                self.get_mock_balance_base(),
                self.get_mock_balance_quote(),
            ]

        @property
        def balance_request_mock_response_only_base(self) -> List[Dict[str, Union[str, int]]]:
            return [self.get_mock_balance_base()]

        def get_mock_balance_base(self) -> Dict[str, Union[str, int]]:
            return {
                "AccountId": self.account_id,
                "Amount": 15,
                "Hold": 5,
                "NotionalHoldAmount": 0,
                "NotionalProductId": 0,
                "NotionalProductSymbol": self.base_asset,
                "NotionalRate": 1,
                "NotionalValue": 0,
                "OMSId": self.oms_id,
                "PendingDeposits": 0,
                "PendingWithdraws": 0,
                "ProductId": self.pair_id + 1,
                "ProductSymbol": self.base_asset,
                "TotalDayDepositNotional": 0,
                "TotalDayDeposits": 0,
                "TotalDayTransferNotional": 0,
                "TotalDayWithdrawNotional": 0,
                "TotalDayWithdraws": 0,
                "TotalMonthDepositNotional": 0,
                "TotalMonthDeposits": 0,
                "TotalMonthWithdrawNotional": 0,
                "TotalMonthWithdraws": 0,
                "TotalYearDepositNotional": 0,
                "TotalYearDeposits": 0,
                "TotalYearWithdrawNotional": 0,
                "TotalYearWithdraws": 0
            }

        def get_mock_balance_quote(self) -> Dict[str, Union[str, int]]:
            return {
                "AccountId": self.account_id,
                "Amount": 2000,
                "Hold": 0,
                "NotionalHoldAmount": 0,
                "NotionalProductId": 0,
                "NotionalProductSymbol": self.quote_asset,
                "NotionalRate": 1,
                "NotionalValue": 0,
                "OMSId": self.oms_id,
                "PendingDeposits": 0,
                "PendingWithdraws": 0,
                "ProductId": self.pair_id + 2,
                "ProductSymbol": self.quote_asset,
                "TotalDayDepositNotional": 0,
                "TotalDayDeposits": 0,
                "TotalDayTransferNotional": 0,
                "TotalDayWithdrawNotional": 0,
                "TotalDayWithdraws": 0,
                "TotalMonthDepositNotional": 0,
                "TotalMonthDeposits": 0,
                "TotalMonthWithdrawNotional": 0,
                "TotalMonthWithdraws": 0,
                "TotalYearDepositNotional": 0,
                "TotalYearDeposits": 0,
                "TotalYearWithdrawNotional": 0,
                "TotalYearWithdraws": 0,
            }

        @property
        def balance_event_websocket_update(self) -> Dict[str, Union[str, int, float]]:
            return {
                "i": 10,
                "m": 3,
                "n": CONSTANTS.WS_ACC_POS_EVENT,
                "o": self.get_mock_balance_base(),
            }

        @property
        def expected_latest_price(self) -> float:
            return 0.390718

        @property
        def expected_supported_order_types(self) -> List[OrderType]:
            return [OrderType.LIMIT]

        @property
        def expected_trading_rule(self):
            trading_rule_resp = self.trading_rules_request_mock_response[0]
            return TradingRule(
                trading_pair=self.trading_pair,
                min_order_size=Decimal(str(trading_rule_resp["MinimumQuantity"])),
                min_price_increment=Decimal(str(trading_rule_resp["PriceIncrement"])),
                min_base_amount_increment=Decimal(str(trading_rule_resp["QuantityIncrement"])),
            )

        @property
        def expected_logged_error_for_erroneous_trading_rule(self):
            erroneous_rule = self.trading_rules_request_erroneous_mock_response[0]
            return f"Error parsing the trading pair rule {erroneous_rule}. Skipping."

        @property
        def expected_exchange_order_id(self):
            return "312269865356374016"

        @property
        def expected_partial_fill_price(self) -> Decimal:
            return Decimal("0.390234")

        @property
        def expected_partial_fill_amount(self) -> Decimal:
            return Decimal("0.5")

        @property
        def expected_fill_fee(self) -> TradeFeeBase:
            return AddedToCostTradeFee(
                percent_token=self.quote_asset,
                flat_fees=[TokenAmount(token=self.quote_asset, amount=Decimal("0.075"))]
            )

        @property
        def expected_fill_trade_id(self) -> int:
            return 123

        @property
        def is_cancel_request_executed_synchronously_by_server(self) -> bool:
            return True

        @property
        def is_order_fill_http_update_included_in_status_update(self) -> bool:
            return True

        @property
        def is_order_fill_http_update_executed_during_websocket_order_event_processing(self) -> bool:
            return False

        def exchange_symbol_for_tokens(self, base_token: str, quote_token: str) -> str:
            return str(self.pair_id)

        def validate_auth_credentials_present(self, request_call: RequestCall):
            request_headers = request_call.kwargs["headers"]
            self.assertEqual(request_headers[CONSTANTS.API_KEY_FIELD], self.api_key)
            self.assertEqual(request_headers[CONSTANTS.SIGNATURE_FIELD], self.signature)
            self.assertEqual(request_headers[CONSTANTS.USER_ID_FIELD], str(self.user_id))
            self.assertEqual(request_headers[CONSTANTS.NONCE_FIELD], self.nonce)

        def validate_order_creation_request(self, order: InFlightOrder, request_call: RequestCall):
            request_data = json.loads(request_call.kwargs["data"])
            self.assertEqual(request_data[CONSTANTS.INSTRUMENT_ID_FIELD], self.pair_id)
            self.assertEqual(request_data[CONSTANTS.OMS_ID_FIELD], self.oms_id)
            self.assertEqual(request_data[CONSTANTS.ACCOUNT_ID_FIELD], self.account_id)
            self.assertEqual(request_data[CONSTANTS.TIME_IN_FORCE_FIELD], CONSTANTS.GTC_TIF)
            self.assertEqual(request_data[CONSTANTS.CLIENT_ORDER_ID_FIELD], int(order.client_order_id))
            self.assertEqual(request_data[CONSTANTS.QUANTITY_FIELD], Decimal("100"))
            self.assertEqual(request_data[CONSTANTS.LIMIT_PRICE_FIELD], Decimal("10000"))
            self.assertEqual(request_data[CONSTANTS.ORDER_TYPE_FIELD], CONSTANTS.LIMIT_ORDER_TYPE)

        def validate_order_status_request(self, order: InFlightOrder, request_call: RequestCall):
            request_data = request_call.kwargs["params"]
            self.assertEqual(request_data[CONSTANTS.OMS_ID_FIELD], self.oms_id)
            self.assertEqual(request_data[CONSTANTS.ACCOUNT_ID_FIELD], self.account_id)
            self.assertEqual(request_data[CONSTANTS.ORDER_ID_FIELD], int(order.exchange_order_id))

        def validate_order_cancelation_request(self, order: InFlightOrder, request_call: RequestCall):
            request_data = json.loads(request_call.kwargs["data"])
            self.assertEqual(request_data[CONSTANTS.OMS_ID_FIELD], self.oms_id)
            self.assertEqual(request_data[CONSTANTS.ACCOUNT_ID_FIELD], self.account_id)
            self.assertEqual(request_data[CONSTANTS.CL_ORDER_ID_FIELD], int(order.client_order_id))

        def validate_trades_request(self, order: InFlightOrder, request_call: RequestCall):
            request_data = request_call.kwargs["params"]
            self.assertEqual(request_data[CONSTANTS.OMS_ID_FIELD], self.oms_id)
            self.assertEqual(request_data[CONSTANTS.ACCOUNT_ID_FIELD], self.account_id)
            self.assertEqual(request_data[CONSTANTS.USER_ID_FIELD], self.user_id)
            self.assertEqual(request_data[CONSTANTS.ORDER_ID_FIELD], int(order.exchange_order_id))

        def configure_successful_cancelation_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_CANCELATION_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = {"detail": None, "errorcode": 0, "errormsg": None, "result": True}
            mock_api.post(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_erroneous_cancelation_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_CANCELATION_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = {"detail": None, "errorcode": 102, "errormsg": "Server Error", "result": False}
            mock_api.post(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_one_successful_one_erroneous_cancel_all_response(
            self,
            successful_order: InFlightOrder,
            erroneous_order: InFlightOrder,
            mock_api: aioresponses,
        ) -> List[str]:
            """
            :return: a list of all configured URLs for the cancelations
            """
            all_urls = []
            url = self.configure_successful_cancelation_response(order=successful_order, mock_api=mock_api)
            all_urls.append(url)
            url = self.configure_erroneous_cancelation_response(order=erroneous_order, mock_api=mock_api)
            all_urls.append(url)
            return all_urls

        def configure_order_not_found_error_cancelation_response(
                self, order: InFlightOrder, mock_api: aioresponses,
                callback: Optional[Callable] = lambda *args, **kwargs: None
        ) -> str:
            # Implement the expected not found response when enabling test_cancel_order_not_found_in_the_exchange
            raise NotImplementedError

        def configure_order_not_found_error_order_status_response(
                self, order: InFlightOrder, mock_api: aioresponses,
                callback: Optional[Callable] = lambda *args, **kwargs: None
        ) -> List[str]:
            # Implement the expected not found response when enabling
            # test_lost_order_removed_if_not_found_during_order_status_update
            raise NotImplementedError

        def configure_completely_filled_order_status_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_STATUS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = self._order_status_request_completely_filled_mock_response(order)
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_canceled_order_status_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_STATUS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = self._order_status_request_canceled_mock_response(order=order)
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_open_order_status_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_STATUS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = self._order_status_request_open_mock_response(order=order)
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_http_error_order_status_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_STATUS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = {"detail": None, "errorcode": 104, "errormsg": "Resource Not Found", "result": False}
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_partially_filled_order_status_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_ORDER_STATUS_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = self._order_status_request_partially_filled_mock_response(order=order)
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_partial_fill_trade_response(
            self,
            order: InFlightOrder,
            mock_api: aioresponses,
            callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_TRADE_HISTORY_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = self._order_fills_request_partial_fill_mock_response(order=order)
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_full_fill_trade_response(
                self,
                order: InFlightOrder,
                mock_api: aioresponses,
                callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_TRADE_HISTORY_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = self._order_fills_request_full_fill_mock_response(order=order)
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def configure_erroneous_http_fill_trade_response(
                self,
                order: InFlightOrder,
                mock_api: aioresponses,
                callback: Optional[Callable] = lambda *args, **kwargs: None,
        ) -> str:
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_TRADE_HISTORY_ENDPOINT)
            regex_url = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            response = {"detail": None, "errorcode": 20, "errormsg": "Not Authorized", "result": False}
            mock_api.get(regex_url, body=json.dumps(response), callback=callback)
            return url

        def order_event_for_new_order_websocket_update(self, order: InFlightOrder) -> Dict:
            return {
                "i": 2,
                "m": 3,
                "n": CONSTANTS.WS_ORDER_STATE_EVENT,
                "o": {
                    "Account": self.account_id,
                    "AccountName": self.user_name,
                    "AvgPrice": 0,
                    "CancelReason": None,
                    "ChangeReason": "NewInputAccepted",
                    "ClientOrderId": int(order.client_order_id),
                    "ClientOrderIdUuid": None,
                    "CounterPartyId": 0,
                    "DisplayQuantity": float(order.amount),
                    "EnteredBy": 169072,
                    "ExecutableValue": 0,
                    "GrossValueExecuted": 0,
                    "InsideAsk": 1455.51,
                    "InsideAskSize": 0.0455,
                    "InsideBid": 1446.28,
                    "InsideBidSize": 1.3325,
                    "Instrument": self.pair_id,
                    "IpAddress": None,
                    "IsLockedIn": False,
                    "IsQuote": False,
                    "LastTradePrice": 1457.34,
                    "LastUpdatedTime": 1655381197195,
                    "LastUpdatedTimeTicks": 637909779971950125,
                    "OMSId": self.oms_id,
                    "OrderFlag": "0",
                    "OrderId": int(order.exchange_order_id),
                    "OrderState": CONSTANTS.ACTIVE_ORDER_STATE,
                    "OrderType": "Limit",
                    "OrigClOrdId": 3,
                    "OrigOrderId": 18846132298,
                    "OrigQuantity": float(order.amount),
                    "PegLimitOffset": 0,
                    "PegOffset": 0,
                    "PegPriceType": "Unknown",
                    "Price": float(order.price),
                    "Quantity": float(order.amount),
                    "QuantityExecuted": 0,
                    "ReceiveTime": 1655381197195,
                    "ReceiveTimeTicks": 637909779971946403,
                    "RejectReason": None,
                    "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                    "StopPrice": 0,
                    "UseMargin": False,
                    "UserName": self.user_name,
                },
            }

        def order_event_for_canceled_order_websocket_update(self, order: InFlightOrder) -> Dict:
            return {
                "i": 6,
                "m": 3,
                "n": CONSTANTS.WS_ORDER_STATE_EVENT,
                "o": {
                    "Account": self.account_id,
                    "AccountName": self.user_name,
                    "AvgPrice": 0,
                    "CancelReason": None,
                    "ChangeReason": "UserModified",
                    "ClientOrderId": int(order.client_order_id),
                    "ClientOrderIdUuid": None,
                    "CounterPartyId": 0,
                    "DisplayQuantity": 0,
                    "EnteredBy": 169072,
                    "ExecutableValue": 0,
                    "GrossValueExecuted": 0,
                    "InsideAsk": 1455.51,
                    "InsideAskSize": 0.0455,
                    "InsideBid": 1446.28,
                    "InsideBidSize": 1.3325,
                    "Instrument": self.pair_id,
                    "IpAddress": None,
                    "IsLockedIn": False,
                    "IsQuote": False,
                    "LastTradePrice": 1457.34,
                    "LastUpdatedTime": 1655381197195,
                    "LastUpdatedTimeTicks": 637909779971950125,
                    "OMSId": self.oms_id,
                    "OrderFlag": "0",
                    "OrderId": int(order.exchange_order_id),
                    "OrderState": CONSTANTS.CANCELED_ORDER_STATE,
                    "OrderType": "Limit",
                    "OrigClOrdId": 3,
                    "OrigOrderId": 18846132298,
                    "OrigQuantity": float(order.amount),
                    "PegLimitOffset": 0,
                    "PegOffset": 0,
                    "PegPriceType": "Unknown",
                    "Price": float(order.price),
                    "Quantity": 0,
                    "QuantityExecuted": 0,
                    "ReceiveTime": 1655381197195,
                    "ReceiveTimeTicks": 637909779971946403,
                    "RejectReason": None,
                    "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                    "StopPrice": 0,
                    "UseMargin": False,
                    "UserName": self.user_name,
                },
            }

        def order_event_for_full_fill_websocket_update(self, order: InFlightOrder) -> Dict:
            return {
                "i": 4,
                "m": 3,
                "n": CONSTANTS.WS_ORDER_TRADE_EVENT,
                "o": self._order_fills_request_full_fill_mock_response(order)[0],
            }

        def trade_event_for_full_fill_websocket_update(self, order: InFlightOrder) -> Dict:
            return {
                "i": 6,
                "m": 3,
                "n": CONSTANTS.WS_ORDER_STATE_EVENT,
                "o": self._order_status_request_completely_filled_mock_response(order),
            }

        def _initialize_auth(self, auth: OMSConnectorAuth):
            auth_resp = self.get_auth_success_response()
            auth.update_with_rest_response(auth_resp)

        def _order_status_request_completely_filled_mock_response(self, order: InFlightOrder) -> Any:
            return {
                "Account": self.account_id,
                "AccountName": self.user_name,
                "AvgPrice": float(order.price),
                "CancelReason": "UserModified",
                "ChangeReason": "UserModified",
                "ClientOrderId": int(order.client_order_id),
                "ClientOrderIdUuid": None,
                "CounterPartyId": 0,
                "DisplayQuantity": 0,
                "EnteredBy": 169072,
                "ExecutableValue": 0.0,
                "GrossValueExecuted": 0.0,
                "InsideAsk": 1452.6,
                "InsideAskSize": 0.0268,
                "InsideBid": 1444.71,
                "InsideBidSize": 0.1185,
                "Instrument": self.pair_id,
                "IpAddress": "85.54.187.233",
                "IsLockedIn": False,
                "IsQuote": False,
                "LastTradePrice": 1444.47,
                "LastUpdatedTime": 1655295880854,
                "LastUpdatedTimeTicks": 637908926808535532,
                "OMSId": self.oms_id,
                "OrderFlag": "AddedToBook, RemovedFromBook",
                "OrderId": order.exchange_order_id,
                "OrderState": CONSTANTS.FULLY_EXECUTED_ORDER_STATE,
                "OrderType": "Limit",
                "OrigClOrdId": 1,
                "OrigOrderId": 18830582877,
                "OrigQuantity": float(order.amount),
                "PegLimitOffset": 0.0,
                "PegOffset": 0.0,
                "PegPriceType": "Unknown",
                "Price": float(order.price),
                "Quantity": 0,
                "QuantityExecuted": float(order.amount),
                "ReceiveTime": 1655295879486,
                "ReceiveTimeTicks": 637908926794855285,
                "RejectReason": None,
                "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                "StopPrice": 0.0,
                "UseMargin": False,
                "UserName": self.user_name,
            }

        def _order_status_request_canceled_mock_response(self, order: InFlightOrder) -> Any:
            return {
                "Account": self.account_id,
                "AccountName": self.user_name,
                "AvgPrice": 0.0,
                "CancelReason": "UserModified",
                "ChangeReason": "UserModified",
                "ClientOrderId": order.client_order_id,
                "ClientOrderIdUuid": None,
                "CounterPartyId": 0,
                "DisplayQuantity": 0,
                "EnteredBy": 169072,
                "ExecutableValue": 0.0,
                "GrossValueExecuted": 0.0,
                "InsideAsk": 1452.6,
                "InsideAskSize": 0.0268,
                "InsideBid": 1444.71,
                "InsideBidSize": 0.1185,
                "Instrument": self.pair_id,
                "IpAddress": "85.54.187.233",
                "IsLockedIn": False,
                "IsQuote": False,
                "LastTradePrice": 1444.47,
                "LastUpdatedTime": 1655295880854,
                "LastUpdatedTimeTicks": 637908926808535532,
                "OMSId": self.oms_id,
                "OrderFlag": "AddedToBook, RemovedFromBook",
                "OrderId": order.exchange_order_id,
                "OrderState": CONSTANTS.CANCELED_ORDER_STATE,
                "OrderType": "Limit",
                "OrigClOrdId": 1,
                "OrigOrderId": 18830582877,
                "OrigQuantity": 0,
                "PegLimitOffset": 0.0,
                "PegOffset": 0.0,
                "PegPriceType": "Unknown",
                "Price": float(order.price),
                "Quantity": float(order.amount),
                "QuantityExecuted": 0.0,
                "ReceiveTime": 1655295879486,
                "ReceiveTimeTicks": 637908926794855285,
                "RejectReason": None,
                "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                "StopPrice": 0.0,
                "UseMargin": False,
                "UserName": self.user_name,
            }

        def _order_status_request_open_mock_response(self, order: InFlightOrder) -> Any:
            return {
                "Account": self.account_id,
                "AccountName": self.user_name,
                "AvgPrice": 0.0,
                "CancelReason": "UserModified",
                "ChangeReason": "UserModified",
                "ClientOrderId": order.client_order_id,
                "ClientOrderIdUuid": None,
                "CounterPartyId": 0,
                "DisplayQuantity": float(order.amount),
                "EnteredBy": 169072,
                "ExecutableValue": 0.0,
                "GrossValueExecuted": 0.0,
                "InsideAsk": 1452.6,
                "InsideAskSize": 0.0268,
                "InsideBid": 1444.71,
                "InsideBidSize": 0.1185,
                "Instrument": self.pair_id,
                "IpAddress": "85.54.187.233",
                "IsLockedIn": False,
                "IsQuote": False,
                "LastTradePrice": 1444.47,
                "LastUpdatedTime": 1655295880854,
                "LastUpdatedTimeTicks": 637908926808535532,
                "OMSId": self.oms_id,
                "OrderFlag": "AddedToBook, RemovedFromBook",
                "OrderId": order.exchange_order_id,
                "OrderState": CONSTANTS.ACTIVE_ORDER_STATE,
                "OrderType": "Limit",
                "OrigClOrdId": 1,
                "OrigOrderId": 18830582877,
                "OrigQuantity": float(order.amount),
                "PegLimitOffset": 0.0,
                "PegOffset": 0.0,
                "PegPriceType": "Unknown",
                "Price": float(order.price),
                "Quantity": float(order.amount),
                "QuantityExecuted": 0.0,
                "ReceiveTime": 1655295879486,
                "ReceiveTimeTicks": 637908926794855285,
                "RejectReason": None,
                "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                "StopPrice": 0.0,
                "UseMargin": False,
                "UserName": self.user_name,
            }

        def _order_status_request_partially_filled_mock_response(self, order: InFlightOrder) -> Any:
            return {
                "Account": self.account_id,
                "AccountName": self.user_name,
                "AvgPrice": float(self.expected_partial_fill_price),
                "CancelReason": "UserModified",
                "ChangeReason": "UserModified",
                "ClientOrderId": order.client_order_id,
                "ClientOrderIdUuid": None,
                "CounterPartyId": 0,
                "DisplayQuantity": float(order.amount),
                "EnteredBy": 169072,
                "ExecutableValue": 0.0,
                "GrossValueExecuted": 0.0,
                "InsideAsk": 1452.6,
                "InsideAskSize": 0.0268,
                "InsideBid": 1444.71,
                "InsideBidSize": 0.1185,
                "Instrument": self.pair_id,
                "IpAddress": "85.54.187.233",
                "IsLockedIn": False,
                "IsQuote": False,
                "LastTradePrice": 1444.47,
                "LastUpdatedTime": 1655295880854,
                "LastUpdatedTimeTicks": 637908926808535532,
                "OMSId": self.oms_id,
                "OrderFlag": "AddedToBook, RemovedFromBook",
                "OrderId": order.exchange_order_id,
                "OrderState": CONSTANTS.ACTIVE_ORDER_STATE,
                "OrderType": "Limit",
                "OrigClOrdId": 1,
                "OrigOrderId": 18830582877,
                "OrigQuantity": float(order.amount),
                "PegLimitOffset": 0.0,
                "PegOffset": 0.0,
                "PegPriceType": "Last",
                "Price": float(order.price),
                "Quantity": float(order.amount),
                "QuantityExecuted": float(self.expected_partial_fill_amount),
                "ReceiveTime": 1655295879486,
                "ReceiveTimeTicks": 637908926794855285,
                "RejectReason": "",
                "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                "StopPrice": 0.0,
                "UseMargin": False,
                "UserName": self.user_name,
            }

        def _order_fills_request_partial_fill_mock_response(self, order: InFlightOrder):
            return [
                {
                    "AccountId": self.account_id,
                    "AccountName": self.user_name,
                    "AdapterTradeId": 0,
                    "ClientOrderId": 0,
                    "CounterParty": "61125",
                    "CounterPartyClientUserId": 1,
                    "Direction": "NoChange",
                    "ExecutionId": 4713579,
                    "Fee": float(self.expected_fill_fee.flat_fees[0].amount),
                    "FeeProductId": self.quote_id,
                    "InsideAsk": 3195.72,
                    "InsideAskSize": 0.0,
                    "InsideBid": 3172.31,
                    "InsideBidSize": 0.0119,
                    "InstrumentId": self.pair_id,
                    "IsBlockTrade": False,
                    "IsQuote": False,
                    "MakerTaker": "Taker",
                    "NotionalHoldAmount": 0,
                    "NotionalProductId": 6,
                    "NotionalRate": 0.8016636123283036,
                    "NotionalValue": 21.34669866907807,
                    "OMSId": self.oms_id,
                    "OrderId": int(order.exchange_order_id),
                    "OrderOriginator": 169072,
                    "OrderTradeRevision": 1,
                    "OrderType": "Market",
                    "Price": float(self.expected_partial_fill_price),
                    "Quantity": float(self.expected_partial_fill_amount),
                    "RemainingQuantity": float(order.amount - self.expected_partial_fill_amount),
                    "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                    "SubAccountId": 0,
                    "TradeId": int(self.expected_fill_trade_id),
                    "TradeTime": 637634811411180462,
                    "TradeTimeMS": 1627884341118,
                    "UserName": self.user_name,
                    "Value": 26.628,
                },
            ]

        def _order_fills_request_full_fill_mock_response(self, order: InFlightOrder):
            return [
                {
                    "AccountId": self.user_id,
                    "AccountName": self.user_name,
                    "AdapterTradeId": 0,
                    "ClientOrderId": int(order.client_order_id),
                    "CounterParty": "61125",
                    "CounterPartyClientUserId": 1,
                    "Direction": "NoChange",
                    "ExecutionId": 4713579,
                    "Fee": float(self.expected_fill_fee.flat_fees[0].amount),
                    "FeeProductId": self.quote_id,
                    "InsideAsk": 3195.72,
                    "InsideAskSize": 0.0,
                    "InsideBid": 3172.31,
                    "InsideBidSize": 0.0119,
                    "InstrumentId": self.pair_id,
                    "IsBlockTrade": False,
                    "IsQuote": False,
                    "MakerTaker": "Taker",
                    "NotionalHoldAmount": 0,
                    "NotionalProductId": 6,
                    "NotionalRate": 0.8016636123283036,
                    "NotionalValue": 21.34669866907807,
                    "OMSId": self.oms_id,
                    "OrderId": int(order.exchange_order_id),
                    "OrderOriginator": 169072,
                    "OrderTradeRevision": 1,
                    "OrderType": "Market",
                    "Price": float(order.price),
                    "Quantity": float(order.amount),
                    "RemainingQuantity": 0.0,
                    "Side": "Buy" if order.trade_type == TradeType.BUY else "Sell",
                    "SubAccountId": 0,
                    "TradeId": int(self.expected_fill_trade_id),
                    "TradeTime": 637634811411180462,
                    "TradeTimeMS": 1627884341118,
                    "UserName": self.user_name,
                    "Value": 26.628,
                },
            ]

        @aioresponses()
        def test_exchange_authenticates(self, mock_api):
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_AUTH_ENDPOINT)
            url_regex = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            auth_resp = self.get_auth_success_response()
            mock_api.get(url_regex, body=json.dumps(auth_resp))

            self.async_run_with_timeout(self.exchange.start_network())

            self.assertTrue(self.exchange.authenticator.initialized)

        @aioresponses()
        def test_failure_to_authenticate(self, mock_api):
            url = self.url_creator.get_rest_url(path_url=CONSTANTS.REST_AUTH_ENDPOINT)
            url_regex = re.compile(f"^{url}".replace(".", r"\.").replace("?", r"\?"))
            auth_resp = self.get_auth_failure_response()
            mock_api.get(url_regex, body=json.dumps(auth_resp))
            exchange = self.create_exchange_instance(authenticated=False)

            with self.assertRaises(IOError):
                self.async_run_with_timeout(exchange.start_network())

            self.assertFalse(exchange.authenticator.initialized)



#!/usr/bin/env python
from typing import (
    Dict,
    Any
)
from sqlalchemy import (
    Column,
    Text,
    Integer,
    BigInteger,
    ForeignKey,
    Index
)
from sqlalchemy.orm import relationship

from . import HummingbotBase


class OrderStatus(HummingbotBase):
    __tablename__ = "OrderStatus"
    __table_args__ = (Index("os_order_id_timestamp_index",
                            "order_id", "timestamp"),
                      )

    id = Column(Integer, primary_key=True, nullable=False)
    order_id = Column(Text, ForeignKey("Order.id"), nullable=False)
    timestamp = Column(BigInteger, nullable=False)
    status = Column(Text, nullable=False)
    order = relationship("Order", back_populates="status")

    def __repr__(self) -> str:
        return f"OrderStatus(id={self.id}, order_id='{self.order_id}', timestamp={self.timestamp}, " \
            f"status='{self.status}')"

    @staticmethod
    def to_bounty_api_json(order_status: "OrderStatus") -> Dict[str, Any]:
        return {
            "order_id": order_status.order_id,
            "timestamp": order_status.timestamp,
            "event_type": order_status.status,
            "raw_json": {
            }
        }



import asyncio
import logging
import math
from decimal import Decimal
from typing import TYPE_CHECKING, Any, AsyncGenerator, AsyncIterable, Dict, Iterable, List, Tuple

from async_timeout import timeout
from bidict import bidict

import hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_constants as constants
import hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_web_utils as web_utils
from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_api_order_book_data_source import (
    CoinbaseAdvancedTradeAPIOrderBookDataSource,
)
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_api_user_stream_data_source import (
    CoinbaseAdvancedTradeAPIUserStreamDataSource,
    CoinbaseAdvancedTradeCumulativeUpdate,
)
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_auth import CoinbaseAdvancedTradeAuth
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_order_book import (
    CoinbaseAdvancedTradeOrderBook,
)
from hummingbot.connector.exchange.coinbase_advanced_trade.coinbase_advanced_trade_web_utils import (
    get_timestamp_from_exchange_time,
    set_exchange_time_from_timestamp,
)
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import TradeFillOrderDetails
from hummingbot.core.data_type.cancellation_result import CancellationResult
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import AddedToCostTradeFee, TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.event.events import MarketEvent, OrderFilledEvent
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter


class CoinbaseAdvancedTradeExchange(ExchangePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 2.5

    web_utils = web_utils

    _logger: HummingbotLogger | logging.Logger | None = None

    @classmethod
    def logger(cls) -> HummingbotLogger | logging.Logger:
        if cls._logger is None:
            name: str = HummingbotLogger.logger_name_for_class(cls)
            cls._logger = logging.getLogger(name)
        return cls._logger

    def __init__(self,
                 client_config_map: "ClientConfigAdapter",
                 coinbase_advanced_trade_api_key: str,
                 coinbase_advanced_trade_api_secret: str,
                 trading_pairs: List[str] | None = None,
                 trading_required: bool = True,
                 domain: str = constants.DEFAULT_DOMAIN,
                 ):
        self._api_key = coinbase_advanced_trade_api_key
        self.secret_key = coinbase_advanced_trade_api_secret
        self._domain = domain
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trades_poll_coinbase_advanced_trade_timestamp = -1
        super().__init__(client_config_map)

        self._asset_uuid_map: Dict[str, str] = {}
        self._pair_symbol_map_initialized = False
        self._market_assets_initialized = False
        self._market_assets: List[Dict[str, Any]] = []

        # Update the time synchronizer logger to the current class logger
        self._time_synchronizer.logger = self.logger
        self.logger().debug(f"{self.name} instance created {self}.")

    def __repr__(self) -> str:
        rep: str = (
            f"CoinbaseAdvancedTradeExchange({self._domain})\n"
            f"  - trading_pairs: {self._trading_pairs}\n"
            f"  - trading_required: {self._trading_required}\n"
            f"  - asset_uuid_map: {self._asset_uuid_map}\n"
            f"  - market_assets_initialized: {self._market_assets_initialized}\n"
            f"  - pair_symbol_map_initialized: {self._market_assets}\n"
            f"  - time_synchronizer: {self._time_synchronizer}\n"
            f"  - last_poll_timestamp: {self._last_poll_timestamp}\n"
            f"  - in_flight_orders: {self._order_tracker.active_orders}\n"
            f"  - status_dict: {self.status_dict}\n"
        )
        return rep

    @property
    def asset_uuid_map(self) -> Dict[str, str]:
        return self._asset_uuid_map

    @staticmethod
    def coinbase_advanced_trade_order_type(order_type: OrderType) -> str:
        if order_type is OrderType.LIMIT_MAKER:
            return "LIMIT_MAKER"
        if order_type is OrderType.LIMIT:
            return "LIMIT"
        if order_type is OrderType.MARKET:
            return "MARKET"

    @staticmethod
    def to_hb_order_type(coinbase_advanced_trade_type: str) -> OrderType:
        return OrderType[coinbase_advanced_trade_type]

    @property
    def authenticator(self):
        return CoinbaseAdvancedTradeAuth(
            api_key=self._api_key,
            secret_key=self.secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        if self._domain == "com":
            return "coinbase_advanced_trade"
        else:
            return f"coinbase_advanced_trade_{self._domain}"

    @property
    def rate_limits_rules(self):
        return constants.RATE_LIMITS

    @property
    def real_time_balance_update(self):
        return False

    @property
    def domain(self):
        return self._domain

    @property
    def last_poll_timestamp(self) -> float:
        # Defined in ExchangePyBase
        return self._last_poll_timestamp

    @property
    def client_order_id_max_length(self):
        return constants.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return constants.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return constants.ALL_PAIRS_EP

    @property
    def trading_pairs_request_path(self):
        return constants.ALL_PAIRS_EP

    @property
    def check_network_request_path(self):
        return constants.SERVER_TIME_EP

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def time_synchronizer(self) -> TimeSynchronizer:
        # Defined in ExchangePyBase
        return self._time_synchronizer

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        # Websocket feed has 'PENDING' state, but REST API does not, it also seems PENDING is for CREATE only
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def in_flight_orders(self) -> Dict[str, InFlightOrder]:
        return self._order_tracker.active_orders

    @property
    def status_dict(self) -> Dict[str, bool]:
        self.logger().debug(
            f"\n   symbols_mapping_initialized: {self.trading_pair_symbol_map_ready()}\n"
            f"   order_books_initialized: {self.order_book_tracker.ready}\n"
            f"   account_balance: {len(self._account_balances) > 0}\n"
            f"   account_balance: {len(self._account_available_balances) > 0}\n"
            f"   trading_required: {self.is_trading_required}\n"
            f"   trading_rule_initialized: {len(self._trading_rules) > 0 if self.is_trading_required else True}\n"
            f"   user_stream_initialized: {self._is_user_stream_initialized()}\n"
        )
        return {
            "symbols_mapping_initialized": self.trading_pair_symbol_map_ready(),
            "order_books_initialized": self.order_book_tracker.ready,
            "account_balance": not self.is_trading_required or len(self._account_balances) > 0,
            "trading_rule_initialized": len(self._trading_rules) > 0 if self.is_trading_required else True,
            "user_stream_initialized": self._is_user_stream_initialized(),
        }

    def supported_order_types(self) -> List[OrderType]:
        return [OrderType.MARKET, OrderType.LIMIT, OrderType.LIMIT_MAKER]

    async def all_trading_pairs(self) -> List[str]:
        """
        List of all trading pairs supported by the connector

        :return: List of trading pair symbols in the Hummingbot format
        """
        # Defined in ExchangeBase, but need to debug!
        mapping = await self.trading_pair_symbol_map()
        return list(mapping.values())

    async def start_network(self):
        await self._initialize_market_assets()
        await self._update_trading_rules()
        self.logger().info("Coinbbase currently not returning trading pairs for USDC in orderbook public messages. setting to USD currently pending fix.")
        await super().start_network()

    def _stop_network(self):
        super()._stop_network()

    async def _update_time_synchronizer(self, pass_on_non_cancelled_error: bool = False):
        # Overriding ExchangePyBase: Synchronizer expects time in ms
        try:
            await self._time_synchronizer.update_server_time_offset_with_time_provider(
                time_provider=self.web_utils.get_current_server_time_ms(
                    throttler=self._throttler,
                    domain=self.domain,
                )
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            if not pass_on_non_cancelled_error:
                self.logger().exception(f"Error requesting time from {self.name_cap} server")
                raise

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        # time endpoint does not communicate an error code
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return CoinbaseAdvancedTradeAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            domain=self.domain,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return CoinbaseAdvancedTradeAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return (
            constants.ORDER_STATUS_NOT_FOUND_ERROR_CODE in str(status_update_exception) or
            "Not Found" in str(status_update_exception) or
            "INVALID_ARGUMENT" in str(status_update_exception)
        )

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return "UNKNOWN_CANCEL_ORDER" in str(cancelation_exception)

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: bool | None = None) -> TradeFeeBase:
        trade_base_fee: TradeFeeBase = build_trade_fee(
            exchange=self.name,
            is_maker=is_maker,
            order_side=order_side,
            order_type=order_type,
            amount=amount,
            price=price,
            base_currency=base_currency,
            quote_currency=quote_currency
        )
        return trade_base_fee

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        """
        Places an order with the exchange and returns the order ID and the timestamp of the order.
        reference: https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder
        Maximum open orders: 500
        """
        amount_str: str = f"{amount:f}"
        price_str: str = f"{price:f}"
        type_str: str = CoinbaseAdvancedTradeExchange.coinbase_advanced_trade_order_type(order_type)
        side_str: str = constants.SIDE_BUY if trade_type is TradeType.BUY else constants.SIDE_SELL
        symbol: str = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        if type_str in {"LIMIT", "LIMIT_MAKER"}:
            order_configuration = {
                "limit_limit_gtc": {
                    "base_size": amount_str,
                    "limit_price": price_str
                }
            }
        # elif type_str == "LIMIT_MAKER":
        #     order_configuration = {
        #         "limit_limit_gtc": {
        #             "base_size": amount_str,
        #             "limit_price": price_str,
        #             # As per HB dev team, LIMIT_MAKER should be converted to LIMIT
        #             # "post_only": True
        #         }
        #     }
        elif type_str == "MARKET":
            if side_str == constants.SIDE_BUY:
                quote_size: Decimal = (amount * price).quantize(
                    self._trading_rules[trading_pair].min_quote_amount_increment)
                order_configuration = {
                    "market_market_ioc": {
                        "quote_size": str(quote_size),
                    }
                }
            else:
                order_configuration = {
                    "market_market_ioc": {
                        "base_size": amount_str,
                    }
                }
        else:
            raise ValueError(f"Invalid order type {order_type}.")

        api_params = {
            "client_order_id": f"{order_id}",
            "product_id": symbol,
            "side": side_str,
            "order_configuration": order_configuration
        }

        order_result = await self._api_post(
            path_url=constants.ORDER_EP,
            data=api_params,
            is_auth_required=True,
        )

        if order_result["success"]:
            o_id = str(order_result["success_response"]["order_id"])
            transact_time = self.time_synchronizer.time()
            self.logger().debug(f"Placed {type_str} order {side_str} {amount_str} {symbol} @ {price_str}")
            return o_id, transact_time

        elif "INSUFFICIENT_FUND" in order_result['error_response']["error"]:
            self.logger().error(
                f"{self.name} reports insufficient funds for {side_str} {amount_str} {symbol} @ {price_str}")
            return "UNKNOWN", self.time_synchronizer.time()

        elif "INVALID_LIMIT_PRICE_POST_ONLY" in order_result['error_response']["error"]:
            self.logger().error(
                f"{self.name} cannot place {type_str} order {side_str} {symbol} @ {price_str}. Likely not POST-able.")
            return "UNKNOWN", self.time_synchronizer.time()

        else:
            raise ValueError(f"Failed to place order on {self.name}. Error: {order_result['error_response']}")

    async def _place_order_and_process_update(self, order: InFlightOrder, **kwargs) -> str:
        # Overriding ExchangePyBase since it sets the status to OPEN
        exchange_order_id, update_timestamp = await self._place_order(
            order_id=order.client_order_id,
            trading_pair=order.trading_pair,
            amount=order.amount,
            trade_type=order.trade_type,
            order_type=order.order_type,
            price=order.price,
            **kwargs,
        )

        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order.client_order_id,
            exchange_order_id=str(exchange_order_id),
            trading_pair=order.trading_pair,
            update_timestamp=update_timestamp,
            new_state=OrderState.PENDING_CREATE if exchange_order_id != "UNKNOWN" else OrderState.FAILED,
        )
        self._order_tracker.process_order_update(order_update)

        return exchange_order_id

    async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:
        """
        Cancels all currently active orders. The cancellations are performed in parallel tasks.

        :param timeout_seconds: the maximum time (in seconds) the cancel logic should run
        :return: a list of CancellationResult instances, one for each of the orders to be cancelled
        """

        async def execute_cancels(order_ids: List[str]) -> List[str]:
            """
            Requests the exchange to cancel an active order

            :param order_ids: the client id of the orders to cancel
            """
            tracked_orders = [(o, t) for o in order_ids if (t := self._order_tracker.fetch_tracked_order(o))]
            result = await self._execute_orders_cancel(orders=[t[1] for t in tracked_orders if t[1] is not None])
            return result

        incomplete_orders = [o for o in self.in_flight_orders.values() if not o.is_done]
        tasks = [execute_cancels([o.client_order_id for o in incomplete_orders])]
        order_id_set = {o.client_order_id for o in incomplete_orders}
        successful_cancellations = []

        try:
            async with timeout(timeout_seconds):
                cancellation_results = await safe_gather(*tasks, return_exceptions=True)
                for task in cancellation_results:
                    if isinstance(task, Exception):
                        continue
                    for client_order_id in task:
                        if client_order_id is not None:
                            order_id_set.remove(client_order_id)
                            successful_cancellations.append(CancellationResult(client_order_id, True))
        except Exception:
            self.logger().network(
                "Unexpected error cancelling orders.",
                exc_info=True,
                app_warning_msg="Failed to cancel order. Check API key and network connection."
            )
        failed_cancellations = [CancellationResult(oid, False) for oid in order_id_set]
        return successful_cancellations + failed_cancellations

    async def _cancel_lost_orders(self):
        await self._execute_orders_cancel(orders=[l for _, l in self._order_tracker.lost_orders.items()])

    async def _execute_orders_cancel(self, orders: List[InFlightOrder]) -> List[str]:
        try:
            cancelled: List[bool] = await self._execute_orders_cancel_and_process_update(orders=orders)
            return [order.client_order_id for order, cancelled in zip(orders, cancelled) if cancelled]

        except asyncio.CancelledError:
            raise

    async def _execute_orders_cancel_and_process_update(self, orders: List[InFlightOrder]) -> List[bool]:
        cancelled = await self._place_cancels(order_ids=[o.exchange_order_id for o in orders])
        for o, c in zip(orders, cancelled):
            if c["success"]:
                update_timestamp = self.current_timestamp
                if update_timestamp is None or math.isnan(update_timestamp):
                    update_timestamp = self.time_synchronizer.time()

                order_update: OrderUpdate = OrderUpdate(
                    client_order_id=o.client_order_id,
                    trading_pair=o.trading_pair,
                    update_timestamp=update_timestamp,
                    new_state=OrderState.CANCELED,
                )
                self._order_tracker.process_order_update(order_update)

            elif c["failure_reason"] in ["UNKNOWN_CANCEL_ORDER", "DUPLICATE_CANCEL_REQUEST"]:
                self.logger().warning(
                    f"Failed to cancel order {o.client_order_id} (order not found OR duplicate request)")
                await self._order_tracker.process_order_not_found(o.client_order_id)
            else:
                self.logger().error(f"Failed to cancel order {o.client_order_id}", exc_info=True)

        return [c["success"] for c in cancelled]

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder) -> bool:
        """
        Cancels an order with the exchange and returns the order ID and the timestamp of the order.
        https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_cancelorders

        :param order_id: str
        :param tracked_order: InFlightOrder
        :return: bool
        """
        # Requesting to cancel an empty order seems to hang the request
        if tracked_order.exchange_order_id is None:
            self.logger().warning(f"Failed to cancel order {order_id} with exchange_id: None")
            self.logger().debug(f"tracked_order: {tracked_order.attributes}")
            return False
        if tracked_order.exchange_order_id == "":
            self.logger().warning(f"Failed to cancel order {order_id} with an empty exchange_id in tracked_order")
            self.logger().debug(f"tracked_order: {tracked_order.attributes}")
            return False
        if tracked_order.exchange_order_id == "UNKNOWN":
            self.logger().error(f"Failed to cancel order {order_id} without exchange_id: UNKNOWN"
                                "File a bug report with the Hummingbot team.")
            raise ValueError(f"Failed to cancel order {order_id} with exchange_id: UNKNOWN")

        result = await self._place_cancels(order_ids=[tracked_order.exchange_order_id])

        if result[0]["success"]:
            return True

        if result[0]["failure_reason"] in ["UNKNOWN_CANCEL_ORDER", "DUPLICATE_CANCEL_REQUEST"]:
            self.logger().warning(f"Failed to cancel order {order_id} (order not found OR duplicate request)")
            await self._order_tracker.process_order_not_found(order_id)

        if result[0]["failure_reason"] in ["UNKNOWN_CANCEL_FAILURE_REASON",
                                           "INVALID_CANCEL_REQUEST",
                                           "COMMANDER_REJECTED_CANCEL_ORDER"]:
            self.logger().error(f"Failed to cancel order {order_id} (Rejected by Coinbase Advanced Trade or Invalid "
                                f"request)")
        return False

    async def _place_cancels(self, order_ids: List[str], max_size: int = 100) -> List[Dict[str, Any]]:
        """
        Cancels an order with the exchange and returns the order ID and the timestamp of the order.
        https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_cancelorders
        MAX_ORDERS is 100 (ChangeLog: 2024-JAN-16)

        :param order_ids: List[str]
        :return: List[Dict[str, Any]]
        """
        # Safeguarding the API call
        order_ids = [o for o in order_ids if o is not None and o != "" and o != "UNKNOWN"]

        if not order_ids:
            return [{"success": True, "failure_reason": "EMPTY_CANCEL_REQUEST"}]

        all_results = []
        for i in range(0, len(order_ids), max_size):
            batched_order_ids = order_ids[i:i + max_size]
            api_data = {
                "order_ids": batched_order_ids
            }
            try:
                cancel_result: Dict[str, Any] = await self._api_post(
                    path_url=constants.BATCH_CANCEL_EP,
                    data=api_data,
                    is_auth_required=True)

                if cancel_result.get("error", False) == 'InvalidArgument':
                    # Error message is 'Too many orderIDs entered, limit is ' + str(NEW_LIMIT)
                    limit = cancel_result.get("message", "").split(" ")[-1]
                    # Resubmit for the remaining orders
                    all_results.extend(await self._place_cancels(order_ids=order_ids[i:], max_size=int(limit)))
                    return all_results

                results: List[Dict[str, Any]] = cancel_result.get("results", [])
                all_results.extend(results)

            except OSError as e:
                self.logger().error(f"Error cancelling orders: {str(e)}\n   {api_data}", exc_info=False)
                return [{"success": False, "failure_reason": "UNKNOWN_CANCEL_FAILURE_REASON"} for _ in order_ids]

        return all_results

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        """
        Get a list of bids/asks for a single product.
        https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getproductbook

        :param trading_pair: str
        :return: OrderBookMessage
        """
        params = {
            "product_id": await self.trading_pair_associated_to_exchange_symbol(trading_pair)
        }

        snapshot: Dict[str, Any] = await self._api_get(
            path_url=constants.SNAPSHOT_EP,
            params=params,
            is_auth_required=True,
            limit_id=constants.SNAPSHOT_EP,
        )

        snapshot_timestamp: float = self.time_synchronizer.time()

        snapshot_msg: OrderBookMessage = CoinbaseAdvancedTradeOrderBook.snapshot_message_from_exchange(
            snapshot,
            snapshot_timestamp,
            metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        """
        Queries Order status by order_id.
        https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_gethistoricalorder

        :param tracked_order: InFlightOrder
        :return: OrderUpdate
        """
        if (
                tracked_order.exchange_order_id is None or
                tracked_order.exchange_order_id == "" or
                tracked_order.exchange_order_id == "UNKNOWN"
        ):
            return OrderUpdate(
                client_order_id=tracked_order.client_order_id,
                exchange_order_id="",
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.time_synchronizer.time(),
                new_state=OrderState.FAILED,
            )

        updated_order_data = await self._api_get(
            path_url=constants.GET_ORDER_STATUS_EP.format(order_id=tracked_order.exchange_order_id),
            params={},
            is_auth_required=True,
            limit_id=constants.GET_ORDER_STATUS_RATE_LIMIT_ID,
        )

        status: str = updated_order_data['order']["status"]
        if status != "UNKNOWN_ORDER_STATUS":
            completion: Decimal = Decimal(updated_order_data['order']["completion_percentage"])
            if status == "OPEN" and completion < Decimal("100"):
                status = "PARTIALLY_FILLED"
        if status not in ["QUEUED", "CANCEL_QUEUED"]:
            new_state = constants.ORDER_STATE[status]

            order_update = OrderUpdate(
                client_order_id=tracked_order.client_order_id,
                exchange_order_id=str(updated_order_data['order']["order_id"]),
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.time_synchronizer.time(),
                new_state=new_state,
            )

            return order_update

    # Overwriting this method from ExchangePyBase that seems to force mis-handling data flow
    # as well as duplicating expensive API calls (call for all products)
    async def _update_trading_rules(self):
        def decimal_or_none(x: Any) -> Decimal | None:
            return Decimal(x) if x is not None else None

        self.trading_rules.clear()
        trading_pair_symbol_map: bidict[str, str] = bidict()

        if not self._market_assets_initialized:
            await self._initialize_market_assets()

        products: List[Dict[str, Any]] = self._market_assets

        if products is None or not products:
            return

        for product in products:
            # Coinbase Advanced Trade API returns the trading pair in the format of "BASE-QUOTE"
            trading_pair: str = product.get("product_id")
            try:
                trading_rule: TradingRule = TradingRule(
                    trading_pair=trading_pair,
                    min_order_size=decimal_or_none(product.get("base_min_size", None)),
                    max_order_size=decimal_or_none(product.get("base_max_size", None)),
                    min_price_increment=decimal_or_none(product.get("quote_increment", None)),
                    min_base_amount_increment=decimal_or_none(product.get("base_increment", None)),
                    min_quote_amount_increment=decimal_or_none(product.get("quote_increment", None)),
                    min_notional_size=decimal_or_none(product.get("quote_min_size", None)),
                    min_order_value=decimal_or_none(product.get("base_min_size", None)) * decimal_or_none(
                        product.get("price", None)),
                    max_price_significant_digits=Decimal(
                        abs(math.floor(
                            math.log10(
                                abs(float(product.get("quote_increment", 0))))))),
                    supports_limit_orders=product.get("supports_limit_orders", False),
                    supports_market_orders=product.get("supports_market_orders", False),
                    buy_order_collateral_token=None,
                    sell_order_collateral_token=None
                )
            except TypeError:
                self.logger().error(
                    f"Error parsing trading pair rule for {product.get('product_id')}, skipping.", exc_info=True,
                )
                continue

            self.trading_rules[trading_pair] = trading_rule

            trading_pair_symbol_map[product.get("product_id", None)] = trading_pair
        self.logger().info("Coinbbase currently not returning trading pairs for USDC in orderbook public messages. setting to USD currently pending fix.")
        self._set_trading_pair_symbol_map(trading_pair_symbol_map)

    async def _initialize_trading_pair_symbol_map(self):
        await self._update_trading_rules()
        self._pair_symbol_map_initialized: bool = True

    async def _initialize_market_assets(self):
        """
        Fetch the list of trading pairs from the exchange and map them
        """
        try:
            params: Dict[str, Any] = {}
            products: Dict[str, Any] = await self._api_get(
                path_url=constants.ALL_PAIRS_EP,
                params=params,
                is_auth_required=True)
            self._market_assets = [p for p in products.get("products") if all((p.get("product_type", None) == "SPOT",
                                                                               p.get("trading_disabled", None) is False,
                                                                               p.get("is_disabled", None) is False,
                                                                               p.get("cancel_only", None) is False,
                                                                               p.get("auction_mode", None) is False))]
            self._market_assets_initialized = True
        except Exception as e:
            self.logger().exception(f"Error getting all trading pairs from Coinbase Advanced Trade: {e}")

    async def _status_polling_loop_fetch_updates(self):
        await self._update_order_fills_from_trades()
        await super()._status_polling_loop_fetch_updates()

    def update_balance(self, asset: str, balance: Decimal):
        self._account_balances[asset] = balance

    def update_available_balance(self, asset: str, balance: Decimal):
        self._account_available_balances[asset] = balance

    def remove_balances(self, assets: Iterable[str]):
        for asset in assets:
            self._account_balances.pop(asset, None)
            self._account_available_balances.pop(asset, None)

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        self.logger().debug("DBG:Balance _update_balance")
        async for account in self._list_trading_accounts():  # type: ignore # Known Pycharm issue
            asset_name: str = account.get("currency")
            hold_value: Decimal = Decimal(account.get("hold").get("value"))
            available_balance: Decimal = Decimal(account.get("available_balance").get("value"))

            # Skip assets with zero balance
            if hold_value == Decimal("0") and available_balance == Decimal("0"):
                continue

            self.update_balance(asset_name, hold_value + available_balance)
            self.update_available_balance(asset_name, available_balance)
            remote_asset_names.add(asset_name)

        # Request removal of non-valid assets
        self.remove_balances(local_asset_names.difference(remote_asset_names))
        self.logger().debug(f"DBG:Balance '-> Balance updated: {self._account_balances}")

    async def _list_one_page_of_accounts(self, cursor: str) -> Dict[str, Any]:
        """
        List one page of accounts with maximum of 250 accounts per page.
        https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getaccounts
        """
        params = {"limit": 250}
        if cursor != "0":
            params["cursor"] = cursor
        response: Dict[str, Any] = await self._api_get(
            path_url=constants.ACCOUNTS_LIST_EP,
            params=params,
            is_auth_required=True,
        )
        return response

    async def _list_trading_accounts(self) -> AsyncGenerator[Dict[str, Any], None]:
        has_next_page = True
        cursor = "0"

        while has_next_page:
            page: Dict[str, Any] = await self._list_one_page_of_accounts(cursor)
            has_next_page = page.get("has_next")
            cursor = page.get("cursor")
            for account in page.get("accounts"):
                self._asset_uuid_map[account.get("currency")] = account.get("uuid")
                yield account

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        product_id = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        params: Dict[str, Any] = {
            "limit": 1,
        }

        trade: Dict[str, Any] = await self._api_get(
            path_url=constants.PAIR_TICKER_24HR_EP.format(product_id=product_id),
            params=params,
            limit_id=constants.PAIR_TICKER_24HR_RATE_LIMIT_ID,
            is_auth_required=True
        )
        return float(trade.get("trades")[0]["price"])

    async def get_all_pairs_prices(self) -> AsyncGenerator[Dict[str, str], None]:
        """
        Fetches the prices of all symbols in the exchange with a default quote of USD
        """
        products: List[Dict[str, str]] = await self._api_get(
            path_url=constants.ALL_PAIRS_EP,
            is_auth_required=True)
        for p in products:
            if all((
                    p.get("product_type", None) == "SPOT",
                    p.get("trading_disabled", None) is False,
                    p.get("is_disabled", None) is False,
                    p.get("cancel_only", None) is False,
                    p.get("auction_mode", None) is False
            )):
                yield {p.get("product_id"): p.get("price")}

    async def get_exchange_rates(self, quote_token: str) -> Dict[str, str] | None:
        """
        Fetches the prices of all symbols in the exchange with a default quote of USD
        """
        response: Dict[str, Any] = await self._api_get(
            path_url=constants.EXCHANGE_RATES_QUOTE_EP.format(quote_token=quote_token),
            limit_id=constants.EXCHANGE_RATES_QUOTE_LIMIT_ID,
            is_auth_required=False)

        data = response.get("data")
        if data is not None and data.get("rates") is not None:
            return data.get("rates")

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        fees: Dict[str, Any] = await self._api_get(path_url=constants.TRANSACTIONS_SUMMARY_EP,
                                                   is_auth_required=True)
        self._trading_fees = fees

    async def _iter_user_event_queue(self) -> AsyncIterable[CoinbaseAdvancedTradeCumulativeUpdate]:
        """
        Called by _user_stream_event_listener.
        """
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Error while reading user events queue. Retrying in 1s.")
                await self._sleep(1.0)

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are order updates.
        """
        async for event_message in self._iter_user_event_queue():
            if isinstance(event_message, dict):
                if (
                        event_message.get("channel") != "user"
                        or event_message.get("sequence_num") != 1
                ):
                    self.logger().error(
                        "Skipping non-cumulative update. This is unintended, but possible, notify devs."
                        f"\n   event_message: {event_message}"
                    )
                continue

            self.logger().debug(f"_user_stream_event_listener: {event_message.client_order_id} {event_message.status}")

            fillable_order: InFlightOrder = self._order_tracker.all_fillable_orders.get(event_message.client_order_id)
            updatable_order: InFlightOrder = self._order_tracker.all_updatable_orders.get(
                event_message.client_order_id)
            state = event_message.status
            if state not in ["QUEUED", "CANCEL_QUEUED"]:
                new_state: OrderState = constants.ORDER_STATE[event_message.status]
                partially: bool = all((event_message.cumulative_base_amount > Decimal("0"),
                                       event_message.remainder_base_amount > Decimal("0"),
                                       new_state == OrderState.OPEN))
                new_state = OrderState.PARTIALLY_FILLED if partially else new_state

                if fillable_order is not None and new_state == OrderState.FILLED:
                    self.logger().debug(
                        f" '-> Fillable: {event_message.client_order_id}. "
                        f"Trigger FILL request at :{self.time_synchronizer.time()}")
                    # This fails the tests, but it is not a problem for the connector
                    # safe_ensure_future(self._update_order_fills_from_trades())
                    await self._update_order_fills_from_trades()

                if updatable_order is not None:
                    self.logger().debug(f" '-> Updatable order: {event_message.client_order_id}")
                    order_update = OrderUpdate(
                        trading_pair=updatable_order.trading_pair,
                        update_timestamp=event_message.fill_timestamp_s,
                        new_state=new_state,
                        client_order_id=event_message.client_order_id,
                        exchange_order_id=event_message.exchange_order_id,
                    )
                    self._order_tracker.process_order_update(order_update)
                else:
                    self.logger().debug(f"Skipping non-updatable order: {event_message.client_order_id}")

    async def _update_order_fills_from_trades(self):
        """
        This is intended to be a backup measure to get filled events with trade ID for orders,
        in case Coinbaase's user stream events are not working.
        NOTE: It is not required to copy this functionality in other connectors.
        This is separated from _update_order_status which only updates the order status without producing filled
        events, since Coinbaase's get order endpoint does not return trade IDs.
        The minimum poll interval for order status is UPDATE_ORDER_STATUS_MIN_INTERVAL seconds.
        """

        def is_execution_time() -> bool:
            """Sets execution time for small and long intervals."""
            small_interval_last_tick: float = self._last_poll_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL
            small_interval_current_tick: float = self.current_timestamp / self.UPDATE_ORDER_STATUS_MIN_INTERVAL
            long_interval_last_tick: float = self._last_poll_timestamp / self.LONG_POLL_INTERVAL
            long_interval_current_tick: float = self.current_timestamp / self.LONG_POLL_INTERVAL

            in_flight_orders: int = len(self.in_flight_orders)

            return (long_interval_current_tick > long_interval_last_tick
                    or (in_flight_orders > 0 and small_interval_current_tick > small_interval_last_tick))

        async def query_trades(pair: str, timestamp=None) -> List[Dict[str, Any]]:
            """Queries trades for a trading pair."""
            trading_pairs = []
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=pair)
            trading_pairs.append(trading_pair)
            p = {"product_ids": trading_pairs}
            if timestamp is not None:
                p["start_sequence_timestamp"] = timestamp

            t: List[Dict[str, Any]] = await self._api_get(
                path_url=constants.FILLS_EP,
                params=p,
                is_auth_required=True)
            return t

        if is_execution_time():
            self.logger().debug(f" '-> Fill request at: {self.time_synchronizer.time()}")
            query_time = set_exchange_time_from_timestamp(self._last_trades_poll_coinbase_advanced_trade_timestamp, "s")
            self._last_trades_poll_coinbase_advanced_trade_timestamp = self.time_synchronizer.time()

            order_by_exchange_id_map = {
                order.exchange_order_id: order
                for order in self._order_tracker.all_fillable_orders.values()
            }

            pairs = self.trading_pairs
            results = await safe_gather(*[query_trades(p, query_time) for p in pairs], return_exceptions=True)

            for trades, trading_pair in zip(results, pairs):
                if isinstance(trades, Exception):
                    self.logger().network(
                        f"Error fetching trades update for the order >{trading_pair}<: >{trades}<.",
                        app_warning_msg=f"Failed to fetch trade update for {trading_pair}."
                    )
                    continue

                for trade in trades["fills"]:
                    exchange_order_id = trade["order_id"]
                    quote_token: str = trading_pair.split("-")[1]

                    if not isinstance(trade["trade_time"], float):
                        trade_time = float(get_timestamp_from_exchange_time(trade["trade_time"], "s"))
                    else:
                        trade_time = trade["trade_time"]

                    f_fee: TokenAmount = TokenAmount(amount=Decimal(trade["commission"]), token=quote_token)
                    fee = AddedToCostTradeFee(percent_token=quote_token, flat_fees=[f_fee])

                    if exchange_order_id in order_by_exchange_id_map:
                        # This is a fill for a tracked order
                        tracked_order = order_by_exchange_id_map[exchange_order_id]

                        if trade["size_in_quote"] is True:
                            fill_quote_amount: Decimal = Decimal(trade["size"])
                            fill_base_amount: Decimal = Decimal(trade["size"]) / Decimal(trade["price"])
                        else:
                            fill_base_amount: Decimal = Decimal(trade["size"])
                            fill_quote_amount: Decimal = Decimal(trade["size"]) * Decimal(trade["price"])

                        trade_update = TradeUpdate(
                            trade_id=str(trade["trade_id"]),
                            client_order_id=tracked_order.client_order_id,
                            exchange_order_id=exchange_order_id,
                            trading_pair=trading_pair,
                            fee=fee,
                            fill_base_amount=fill_base_amount,
                            fill_quote_amount=fill_quote_amount,
                            fill_price=Decimal(trade["price"]),
                            fill_timestamp=trade_time,
                            is_taker=False
                        )
                        self._order_tracker.process_trade_update(trade_update)

                    elif self.is_confirmed_new_order_filled_event(str(trade["trade_id"]),
                                                                  str(exchange_order_id),
                                                                  trading_pair):
                        # This is a fill of an order registered in the DB but not tracked anymore
                        self._current_trade_fills.add(TradeFillOrderDetails(
                            market=self.display_name,
                            exchange_trade_id=str(trade["trade_id"]),
                            symbol=trading_pair))
                        self.trigger_event(
                            MarketEvent.OrderFilled,
                            OrderFilledEvent(
                                timestamp=trade_time,
                                order_id=self._exchange_order_ids.get(str(trade["order_id"]), None),
                                trading_pair=trading_pair,
                                trade_type=TradeType.BUY if trade["side"] == "BUY" else TradeType.SELL,
                                order_type=OrderType.LIMIT,
                                price=Decimal(trade["price"]),
                                amount=Decimal(trade["size"]),
                                trade_fee=fee,
                                exchange_trade_id=str(trade["trade_id"])
                            ))
                        self.logger().info(
                            f"Recreating missing trade {trade['side']} {trade['size']} {trading_pair} @ {trade['price']}")
                    else:
                        self.logger().debug(f"Trade without matching order_id and not in the DB: {trade}")

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        """
        Queries all trades for an order.
        https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getfills
        """
        trade_updates = []
        if order.exchange_order_id is not None:
            order_ids = []
            order_id: str = order.exchange_order_id
            order_ids.append(str(order_id))
            # product_id: str = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            params = {
                "order_ids": order_ids
            }
            all_fills_response: Dict[str, Any] = await self._api_get(
                path_url=constants.FILLS_EP,
                params=params,
                is_auth_required=True)

            for trade in all_fills_response["fills"]:
                exchange_order_id = trade["order_id"]
                quote_token: str = order.trading_pair.split("-")[1]
                if not isinstance(trade["trade_time"], float):
                    trade_time = float(get_timestamp_from_exchange_time(trade["trade_time"], "s"))
                else:
                    trade_time = trade["time"]
                fee = AddedToCostTradeFee(
                    percent_token=quote_token,
                    flat_fees=[TokenAmount(
                        amount=Decimal(trade["commission"]),
                        token=quote_token)])
                if trade["size_in_quote"] is True:
                    fill_quote_amount: Decimal = Decimal(trade["size"])
                    fill_base_amount: Decimal = Decimal(trade["size"]) / Decimal(trade["price"])
                else:
                    fill_base_amount: Decimal = Decimal(trade["size"])
                    fill_quote_amount: Decimal = Decimal(trade["size"]) * Decimal(trade["price"])
                trade_update = TradeUpdate(
                    trade_id=str(trade["trade_id"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=exchange_order_id,
                    trading_pair=order.trading_pair,
                    fee=fee,
                    fill_base_amount=fill_base_amount,
                    fill_quote_amount=fill_quote_amount,
                    fill_price=Decimal(trade["price"]),
                    fill_timestamp=trade_time,
                )
                trade_updates.append(trade_update)

        return trade_updates

    async def _api_post(self, *args, **kwargs):
        return await super()._api_post(*args, **kwargs)

    async def _api_get(self, *args, **kwargs):
        return await super()._api_get(*args, **kwargs)

    async def _make_network_check_request(self):
        self.logger().debug(f"Checking network status of {self.name} by querying server time.")
        await self._api_get(path_url=constants.SERVER_TIME_EP)

    async def _format_trading_rules(self, e: Dict[str, Any]) -> List[TradingRule]:
        raise NotImplementedError(f"This method is not implemented by {self.name} connector")

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        raise NotImplementedError(f"This method is not implemented by {self.name} connector")

    def _make_trading_rules_request(self) -> Any:
        raise NotImplementedError(f"This method is not implemented by {self.name} connector")

    def _make_trading_pairs_request(self) -> Any:
        raise NotImplementedError(f"This method is not implemented by {self.name} connector")
